##########################################################################
#
# Processor specific code

# CPU = "6811"
# Description = "FreeScale 68HC11 8-bit microcontroller."
# DataWidth = 8 # 8-bit data
# AddressWidth = 16 # 16-bit addresses

# Notes:

# 1. Some instructions use the same opcode (e.g. bcs and blo).
#    Disassembly will use whatever is listed last in this table.
# 2. brset/brclr instructions do not yet output correct branch address.

# Maximum length of an instruction (for formatting purposes)
maxLength = 5

# Leadin bytes for multbyte instructions
leadInBytes = [0x18, 0x1a, 0xcd]

# Addressing mode table
addressModeTable = {
"inherent"   : "",
"immediate"  : "#${0:02X}",
"immediatex" : "#${0:02X}{1:02X}",
"direct"     : "${0:02X}",
"direct2"    : "*${0:02X} ${1:02X}",
"direct3"    : "*${0:02X} ${1:02X} ${2:04X}",
"extended"   : "${0:02X}{1:02X}",
"indexedx"   : "${0:02X},x",
"indexedx2"  : "${0:02X},x ${1:02X}",
"indexedx3"  : "${0:02X},x ${1:02X} ${2:04X}",
"indexedy"   : "${0:02X},y",
"indexedy2"  : "${0:02X},y ${1:02X}",
"indexedy3"  : "${0:02X},y ${1:02X} ${2:04X}",
"relative"   : "${0:04X}",
}

# Op Code Table
# Key is numeric opcode (possibly multiple bytes)
# Value is a list:
#   # bytes
#   mnemonic
#   addressing mode.
#   flags (e.g. pcr)
opcodeTable = {

0x1b   :  [ 1, "aba",  "inherent"        ],
0x3a   :  [ 1, "abx",  "inherent"        ],
0x183a :  [ 2, "aby",  "inherent"        ],
0x89   :  [ 2, "adca", "immediate"       ],
0x99   :  [ 2, "adca", "direct"          ],
0xb9   :  [ 3, "adca", "extended"        ],
0xa9   :  [ 2, "adca", "indexedx"        ],
0x18a9 :  [ 3, "adca", "indexedy"        ],
0xc9   :  [ 2, "adcb", "immediate"       ],
0xd9   :  [ 2, "adcb", "direct"          ],
0xf9   :  [ 3, "adcb", "extended"        ],
0xe9   :  [ 2, "adcb", "indexedx"        ],
0x18e9 :  [ 3, "adcb", "indexedy"        ],
0x8b   :  [ 2, "adda", "immediate"       ],
0x9b   :  [ 2, "adda", "direct"          ],
0xbb   :  [ 3, "adda", "extended"        ],
0xab   :  [ 2, "adda", "indexedx"        ],
0x18ab :  [ 3, "adda", "indexedy"        ],
0xcb   :  [ 2, "addb", "immediate"       ],
0xdb   :  [ 2, "addb", "direct"          ],
0xfb   :  [ 3, "addb", "extended"        ],
0xeb   :  [ 2, "addb", "indexedx"        ],
0x18eb :  [ 3, "addb", "indexedy"        ],
0xc3   :  [ 3, "addd", "immediatex"      ],
0xd3   :  [ 2, "addd", "direct"          ],
0xf3   :  [ 3, "addd", "extended"        ],
0xe3   :  [ 2, "addd", "indexedx"        ],
0x18e3 :  [ 3, "addd", "indexedy"        ],
0x84   :  [ 2, "anda", "immediate"       ],
0x94   :  [ 2, "anda", "direct"          ],
0xb4   :  [ 3, "anda", "extended"        ],
0xa4   :  [ 2, "anda", "indexedx"        ],
0x18a4 :  [ 3, "anda", "indexedy"        ],
0xc4   :  [ 2, "andb", "immediate"       ],
0xd4   :  [ 2, "andb", "direct"          ],
0xf4   :  [ 3, "andb", "extended"        ],
0xe4   :  [ 2, "andb", "indexedx"        ],
0x18e4 :  [ 3, "andb", "indexedy"        ],
0x78   :  [ 3, "asl",  "extended"        ],
0x68   :  [ 2, "asl",  "indexedx"        ],
0x1868 :  [ 3, "asl",  "indexedy"        ],
0x48   :  [ 1, "asla", "inherent"        ],
0x58   :  [ 1, "aslb", "inherent"        ],
0x05   :  [ 1, "asld", "inherent"        ],
0x77   :  [ 3, "asr",  "extended"        ],
0x67   :  [ 2, "asr",  "indexedx"        ],
0x1867 :  [ 3, "asr",  "indexedy"        ],
0x47   :  [ 1, "asra", "inherent"        ],
0x57   :  [ 1, "asrb", "inherent"        ],
0x24   :  [ 2, "bcc",  "relative", pcr   ],
0x15   :  [ 3, "bclr", "direct2",        ],
0x1d   :  [ 3, "bclr", "indexedx2",      ],
0x181d :  [ 4, "bclr", "indexedy2",      ],
0x25   :  [ 2, "bcs",  "relative", pcr   ],
0x27   :  [ 2, "beq",  "relative", pcr   ],
0x2c   :  [ 2, "bge",  "relative", pcr   ],
0x2e   :  [ 2, "bgt",  "relative", pcr   ],
0x22   :  [ 2, "bhi",  "relative", pcr   ],
0x24   :  [ 2, "bhs",  "relative", pcr   ],
0x85   :  [ 2, "bita", "immediate"       ],
0x95   :  [ 2, "bita", "direct"          ],
0xb5   :  [ 3, "bita", "extended"        ],
0xa5   :  [ 2, "bita", "indexedx"        ],
0x18a5 :  [ 3, "bita", "indexedy"        ],
0xc5   :  [ 2, "bitb", "immediate"       ],
0xd5   :  [ 2, "bitb", "direct"          ],
0xf5   :  [ 3, "bitb", "extended"        ],
0xe5   :  [ 2, "bitb", "indexedx"        ],
0x18e5 :  [ 3, "bitb", "indexedy"        ],
0x2f   :  [ 2, "ble",  "relative", pcr   ],
0x25   :  [ 2, "blo",  "relative", pcr   ],
0x23   :  [ 2, "bls",  "relative", pcr   ],
0x2d   :  [ 2, "blt",  "relative", pcr   ],
0x2b   :  [ 2, "bmi",  "relative", pcr   ],
0x26   :  [ 2, "bne",  "relative", pcr   ],
0x2a   :  [ 2, "bpl",  "relative", pcr   ],
0x20   :  [ 2, "bra",  "relative", pcr   ],
0x13   :  [ 4, "brclr", "direct3", pcr   ],
0x1f   :  [ 4, "brclr", "indexedx3", pcr ],
0x181f :  [ 5, "brclr", "indexedy3", pcr ],
0x21   :  [ 2, "brn",  "relative", pcr   ],
0x12   :  [ 4, "brset", "direct3", pcr   ],
0x1e   :  [ 4, "brset", "indexedx3", pcr ],
0x181e :  [ 5, "brset", "indexedy3", pcr ],
0x14   :  [ 3, "bset", "direct2",        ],
0x1c   :  [ 3, "bset", "indexedx2",      ],
0x181c :  [ 4, "bset", "indexedy2",      ],
0x8d   :  [ 2, "bsr",  "relative", pcr   ],
0x28   :  [ 2, "bvc",  "relative", pcr   ],
0x29   :  [ 2, "bvs",  "relative", pcr   ],
0x11   :  [ 1, "cba",  "inherent"        ],
0x0c   :  [ 1, "clc",  "inherent"        ],
0x0e   :  [ 1, "cli",  "inherent"        ],
0x7f   :  [ 3, "clr",  "extended"        ],
0x6f   :  [ 2, "clr",  "indexedx"        ],
0x187f :  [ 3, "clr",  "indexedy"        ],
0x4f   :  [ 1, "clra", "inherent"        ],
0x5f   :  [ 1, "clrb", "inherent"        ],
0x0a   :  [ 1, "clv",  "inherent"        ],
0x81   :  [ 2, "cmpa", "immediate"       ],
0x91   :  [ 2, "cmpa", "direct"          ],
0xb1   :  [ 3, "cmpa", "extended"        ],
0xa1   :  [ 2, "cmpa", "indexedx"        ],
0x18a1 :  [ 3, "cmpa", "indexedy"        ],
0xc1   :  [ 2, "cmpb", "immediate"       ],
0xd1   :  [ 2, "cmpb", "direct"          ],
0xf1   :  [ 3, "cmpb", "extended"        ],
0xe1   :  [ 2, "cmpb", "indexedx"        ],
0x18e1 :  [ 3, "cmpb", "indexedy"        ],
0x73   :  [ 3, "com",  "extended"        ],
0x63   :  [ 2, "com",  "indexedx"        ],
0x1863 :  [ 3, "com",  "indexedy"        ],
0x43   :  [ 1, "coma", "inherent"        ],
0x53   :  [ 1, "comb", "inherent"        ],
0x1a83 :  [ 4, "cpd",  "immediatex"      ],
0x1a93 :  [ 3, "cpd",  "direct"          ],
0x1ab3 :  [ 4, "cpd",  "extended"        ],
0x1aa3 :  [ 3, "cpd",  "indexedx"        ],
0xcda3 :  [ 3, "cpd",  "indexedy"        ],
0x8c   :  [ 3, "cpx",  "immediatex"      ],
0x9c   :  [ 2, "cpx",  "direct"          ],
0xbc   :  [ 3, "cpx",  "extended"        ],
0xac   :  [ 2, "cpx",  "indexedx"        ],
0xcdac :  [ 3, "cpx",  "indexedy"        ],
0x188c :  [ 4, "cpy",  "immediatex"      ],
0x189c :  [ 3, "cpy",  "direct"          ],
0x18bc :  [ 4, "cpy",  "extended"        ],
0x1aac :  [ 3, "cpy",  "indexedx"        ],
0x18ac :  [ 3, "cpy",  "indexedy"        ],
0x19   :  [ 1, "daa",  "inherent"        ],
0x7a   :  [ 3, "dec",  "extended"        ],
0x6a   :  [ 2, "dec",  "indexedx"        ],
0x186a :  [ 3, "dec",  "indexedy"        ],
0x4a   :  [ 1, "deca", "inherent"        ],
0x5a   :  [ 1, "decb", "inherent"        ],
0x34   :  [ 1, "des",  "inherent"        ],
0x09   :  [ 1, "dex",  "inherent"        ],
0x1809 :  [ 2, "dey",  "inherent"        ],
0x88   :  [ 2, "eora", "immediate"       ],
0x98   :  [ 2, "eora", "direct"          ],
0xb8   :  [ 3, "eora", "extended"        ],
0xa8   :  [ 2, "eora", "indexedx"        ],
0x18a8 :  [ 3, "eora", "indexedy"        ],
0xc8   :  [ 2, "eorb", "immediate"       ],
0xd8   :  [ 2, "eorb", "direct"          ],
0xf8   :  [ 3, "eorb", "extended"        ],
0xe8   :  [ 2, "eorb", "indexedx"        ],
0x18e8 :  [ 3, "eorb", "indexedy"        ],
0x03   :  [ 1, "fdiv", "inherent"        ],
0x02   :  [ 1, "idiv", "inherent"        ],
0x7c   :  [ 3, "inc",  "extended"        ],
0x6c   :  [ 2, "inc",  "indexedx"        ],
0x186c :  [ 3, "inc",  "indexedy"        ],
0x4c   :  [ 1, "inca", "inherent"        ],
0x5c   :  [ 1, "incb", "inherent"        ],
0x31   :  [ 1, "ins",  "inherent"        ],
0x08   :  [ 1, "inx",  "inherent"        ],
0x1808 :  [ 2, "iny",  "inherent"        ],
0x7e   :  [ 3, "jmp",  "extended"        ],
0x6e   :  [ 2, "jmp",  "indexedx"        ],
0x186e :  [ 3, "jmp",  "indexedy"        ],
0x9d   :  [ 2, "jsr",  "direct"          ],
0xbd   :  [ 3, "jsr",  "extended"        ],
0xad   :  [ 2, "jsr",  "indexedx"        ],
0x18ad :  [ 3, "jsr",  "indexedy"        ],
0x86   :  [ 2, "ldaa", "immediate"       ],
0x96   :  [ 2, "ldaa", "direct"          ],
0xb6   :  [ 3, "ldaa", "extended"        ],
0xa6   :  [ 2, "ldaa", "indexedx"        ],
0x18a6 :  [ 3, "ldaa", "indexedy"        ],
0xc6   :  [ 2, "ldab", "immediate"       ],
0xd6   :  [ 2, "ldab", "direct"          ],
0xf6   :  [ 3, "ldab", "extended"        ],
0xe6   :  [ 2, "ldab", "indexedx"        ],
0x18e6 :  [ 3, "ldab", "indexedy"        ],
0xcc   :  [ 3, "ldd",  "immediatex"      ],
0xdc   :  [ 2, "ldd",  "direct"          ],
0xfc   :  [ 3, "ldd",  "extended"        ],
0xec   :  [ 2, "ldd",  "indexedx"        ],
0x18ec :  [ 3, "ldd",  "indexedy"        ],
0x8e   :  [ 3, "lds",  "immediatex"      ],
0x9e   :  [ 2, "lds",  "direct"          ],
0xbe   :  [ 3, "lds",  "extended"        ],
0xae   :  [ 2, "lds",  "indexedx"        ],
0x18ae :  [ 3, "lds",  "indexedy"        ],
0xce   :  [ 3, "ldx",  "immediatex"      ],
0xde   :  [ 2, "ldx",  "direct"          ],
0xfe   :  [ 3, "ldx",  "extended"        ],
0xee   :  [ 2, "ldx",  "indexedx"        ],
0xcdee :  [ 3, "ldx",  "indexedy"        ],
0x18ce :  [ 4, "ldy",  "immediatex"      ],
0x18de :  [ 3, "ldy",  "direct"          ],
0x18fe :  [ 4, "ldy",  "extended"        ],
0x1aee :  [ 3, "ldy",  "indexedx"        ],
0x18ee :  [ 3, "ldy",  "indexedy"        ],
0x78   :  [ 3, "lsl",  "extended"        ],
0x68   :  [ 2, "lsl",  "indexedx"        ],
0x1868 :  [ 3, "lsl",  "indexedy"        ],
0x48   :  [ 1, "lsla", "inherent"        ],
0x58   :  [ 1, "lslb", "inherent"        ],
0x05   :  [ 1, "lsld", "inherent"        ],
0x74   :  [ 3, "lsr",  "extended"        ],
0x64   :  [ 2, "lsr",  "indexedx"        ],
0x1864 :  [ 3, "lsr",  "indexedy"        ],
0x44   :  [ 1, "lsra", "inherent"        ],
0x54   :  [ 1, "lsrb", "inherent"        ],
0x04   :  [ 1, "lsrd", "inherent"        ],
0x3d   :  [ 1, "mul",  "inherent"        ],
0x70   :  [ 3, "neg",  "extended"        ],
0x60   :  [ 2, "neg",  "indexedx"        ],
0x1860 :  [ 3, "neg",  "indexedy"        ],
0x40   :  [ 3, "nega", "inherent"        ],
0x50   :  [ 3, "negb", "inherent"        ],
0x01   :  [ 1, "nop",  "inherent"        ],
0x8a   :  [ 2, "oraa", "immediate"       ],
0x9a   :  [ 2, "oraa", "direct"          ],
0xba   :  [ 3, "oraa", "extended"        ],
0xaa   :  [ 2, "oraa", "indexedx"        ],
0x18aa :  [ 3, "oraa", "indexedy"        ],
0xca   :  [ 2, "orab", "immediate"       ],
0xda   :  [ 2, "orab", "direct"          ],
0xfa   :  [ 3, "orab", "extended"        ],
0xea   :  [ 2, "orab", "indexedx"        ],
0x18ea :  [ 3, "orab", "indexedy"        ],
0x36   :  [ 1, "psha", "inherent"        ],
0x37   :  [ 1, "pshb", "inherent"        ],
0x3c   :  [ 1, "pshx", "inherent"        ],
0x183c :  [ 2, "pshy", "inherent"        ],
0x32   :  [ 1, "pula", "inherent"        ],
0x33   :  [ 1, "pulb", "inherent"        ],
0x38   :  [ 1, "pulx", "inherent"        ],
0x1838 :  [ 2, "puly", "inherent"        ],
0x79   :  [ 3, "rol",  "extended"        ],
0x69   :  [ 2, "rol",  "indexedx"        ],
0x1869 :  [ 3, "rol",  "indexedy"        ],
0x49   :  [ 1, "rola", "inherent"        ],
0x59   :  [ 1, "rolb", "inherent"        ],
0x76   :  [ 3, "ror",  "extended"        ],
0x66   :  [ 2, "ror",  "indexedx"        ],
0x1866 :  [ 3, "ror",  "indexedy"        ],
0x46   :  [ 1, "rora", "inherent"        ],
0x55   :  [ 1, "rorb", "inherent"        ],
0x3b   :  [ 1, "rti",  "inherent"        ],
0x39   :  [ 1, "rts",  "inherent"        ],
0x10   :  [ 1, "sba",  "inherent"        ],
0x82   :  [ 2, "sbca", "immediate"       ],
0x92   :  [ 2, "sbca", "direct"          ],
0xb2   :  [ 3, "sbca", "extended"        ],
0xa2   :  [ 2, "sbca", "indexedx"        ],
0x18a2 :  [ 3, "sbca", "indexedy"        ],
0xc2   :  [ 2, "sbcb", "immediate"       ],
0xd2   :  [ 2, "sbcb", "direct"          ],
0xf2   :  [ 3, "sbcb", "extended"        ],
0xe2   :  [ 2, "sbcb", "indexedx"        ],
0x18e2 :  [ 3, "sbcb", "indexedy"        ],
0x0d   :  [ 1, "sec",  "inherent"        ],
0x0f   :  [ 1, "sei",  "inherent"        ],
0x0b   :  [ 1, "sev",  "inherent"        ],
0x97   :  [ 2, "staa", "direct"          ],
0xb7   :  [ 3, "staa", "extended"        ],
0xa7   :  [ 2, "staa", "indexedx"        ],
0x18a7 :  [ 3, "staa", "indexedy"        ],
0xd7   :  [ 2, "stab", "direct"          ],
0xf7   :  [ 3, "stab", "extended"        ],
0xe7   :  [ 2, "stab", "indexedx"        ],
0x18e7 :  [ 3, "stab", "indexedy"        ],
0xdd   :  [ 2, "std",  "direct"          ],
0xfd   :  [ 3, "std",  "extended"        ],
0xed   :  [ 2, "std",  "indexedx"        ],
0x18ed :  [ 3, "std",  "indexedy"        ],
0xcf   :  [ 1, "stop", "inherent"        ],
0x9f   :  [ 2, "sts",  "direct"          ],
0xbf   :  [ 3, "sts",  "extended"        ],
0xaf   :  [ 2, "sts",  "indexedx"        ],
0x18af :  [ 3, "sts",  "indexedy"        ],
0xdf   :  [ 2, "stx",  "direct"          ],
0xff   :  [ 3, "stx",  "extended"        ],
0xef   :  [ 2, "stx",  "indexedx"        ],
0xcdef :  [ 3, "stx",  "indexedy"        ],
0x18df :  [ 3, "sty",  "direct"          ],
0x18ff :  [ 4, "sty",  "extended"        ],
0x1aef :  [ 3, "sty",  "indexedx"        ],
0x18ef :  [ 3, "sty",  "indexedy"        ],
0x80   :  [ 2, "suba", "immediate"       ],
0x90   :  [ 2, "suba", "direct"          ],
0xb0   :  [ 3, "suba", "extended"        ],
0xa0   :  [ 2, "suba", "indexedx"        ],
0x18a0 :  [ 3, "suba", "indexedy"        ],
0xc0   :  [ 2, "subb", "immediate"       ],
0xd0   :  [ 2, "subb", "direct"          ],
0xf0   :  [ 3, "subb", "extended"        ],
0xe0   :  [ 2, "subb", "indexedx"        ],
0x18e0 :  [ 3, "subb", "indexedy"        ],
0x83   :  [ 3, "subd", "immediatex"      ],
0x93   :  [ 2, "subd", "direct"          ],
0xb3   :  [ 3, "subd", "extended"        ],
0xa3   :  [ 2, "subd", "indexedx"        ],
0x18a3 :  [ 3, "subd", "indexedy"        ],
0x3f   :  [ 1, "swi",  "inherent"        ],
0x16   :  [ 1, "tab",  "inherent"        ],
0x06   :  [ 1, "tap",  "inherent"        ],
0x17   :  [ 1, "tba",  "inherent"        ],
0x00   :  [ 1, "test", "inherent"        ],
0x07   :  [ 1, "tpa",  "inherent"        ],
0x7d   :  [ 3, "tst",  "extended"        ],
0x6d   :  [ 2, "tst",  "indexedx"        ],
0x186d :  [ 3, "tst",  "indexedy"        ],
0x4d   :  [ 1, "tsta",  "inherent"       ],
0x5d   :  [ 1, "tstb",  "inherent"       ],
0x30   :  [ 1, "tsx",   "inherent"       ],
0x1830 :  [ 2, "tsy",   "inherent"       ],
0x35   :  [ 1, "txs",   "inherent"       ],
0x1835 :  [ 2, "tys",   "inherent"       ],
0x3e   :  [ 1, "wai",   "inherent"       ],
0x8f   :  [ 1, "xgdx", "inherent"        ],
0x188f :  [ 2, "xgdy", "inherent"        ],

}

# End of processor specific code
##########################################################################



import os

from traits.trait_base import get_resource_path

def get_template(name):
    path = get_resource_path(1)
    print path, name
    pathname = os.path.normpath("%s/%s" % (path, name))
    if os.path.exists(pathname):
        with open(pathname, "rb") as fh:
            source = fh.read()
        return source



#-------------------------------------------------------------------------------
#
#  Copyright (c) 2007, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   06/22/2007
#
#-------------------------------------------------------------------------------

""" Traits UI simple, themed slider-based integer or float value editor.
"""

#-------------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------------

import wx

from math \
   import log10, pow

from traits.api \
    import HasPrivateTraits, Instance, Enum, Range, Str, Float, Bool, Color, \
           TraitError

from traitsui.ui_traits \
    import Alignment

from traitsui.wx.editor \
    import Editor

from traitsui.basic_editor_factory \
    import BasicEditorFactory

from pyface.timer.api \
    import do_after

from constants \
    import ErrorColor

from helper \
    import disconnect, disconnect_no_id, BufferDC

#-------------------------------------------------------------------------------
#  '_ThemedSliderEditor' class:
#-------------------------------------------------------------------------------

class _ThemedSliderEditor ( Editor ):
    """ Traits UI simple, themed slider-based integer or float value editor.
    """

    # The low end of the slider range:
    low = Float

    # The high end of the slider range:
    high = Float

    # The smallest allowed increment:
    increment = Float

    # The current text being displayed:
    text = Str

    #-- Class Variables --------------------------------------------------------

    text_styles = {
        'left':   wx.TE_LEFT,
        'center': wx.TE_CENTRE,
        'right':  wx.TE_RIGHT
    }

    #---------------------------------------------------------------------------
    #  Finishes initializing the editor by creating the underlying toolkit
    #  widget:
    #---------------------------------------------------------------------------

    def init ( self, parent ):
        """ Finishes initializing the editor by creating the underlying toolkit
            widget.
        """
        factory = self.factory
        # Establish the range of the slider:
        low, high = factory.low, factory.high
        if high <= low:
            low = high = None
            range      = self.object.trait( self.name ).handler
            if isinstance( range, Range ):
                low, high = range._low, range._high
            if low is None:
                if high is None:
                    high = 1.0
                low = high - 1.0
            elif high is None:
                high = low + 1.0

        # Establish the slider increment:
        increment = factory.increment
        if increment <= 0.0:
            if isinstance( low, int ):
                increment = 1.0
            else:
                increment = pow( 10, int( log10( (high - low) / 1000.0 ) ) )

        # Save the values we calculated:
        self.set( low = low, high = high, increment = increment )

        # Create the control:
        self.control = control = wx.Window( parent, -1,
                                            size  = wx.Size( 70, 20 ),
                                            style = wx.FULL_REPAINT_ON_RESIZE |
                                                    wx.TAB_TRAVERSAL )

        # Set up the painting event handlers:
        wx.EVT_ERASE_BACKGROUND( control, self._erase_background )
        wx.EVT_PAINT( control, self._on_paint )
        wx.EVT_SET_FOCUS(  control, self._set_focus )

        # Set up mouse event handlers:
        wx.EVT_LEFT_DOWN(    control, self._left_down )
        wx.EVT_LEFT_UP(      control, self._left_up )
        wx.EVT_MOTION(       control, self._motion )
        wx.EVT_MOUSEWHEEL(   control, self._mouse_wheel )
        wx.EVT_ENTER_WINDOW( control, self._enter_window )
        wx.EVT_LEAVE_WINDOW( control, self._leave_window )

        # Set up the control resize handler:
        wx.EVT_SIZE( control, self._resize )

        # Set the tooltip:
        if not self.set_tooltip():
            control.SetToolTipString( '[%g..%g]' % ( low, high ) )

    #---------------------------------------------------------------------------
    #  Disposes of the contents of an editor:
    #---------------------------------------------------------------------------

    def dispose ( self ):
        """ Disposes of the contents of an editor.
        """
        disconnect_no_id( self.control, wx.EVT_ERASE_BACKGROUND, wx.EVT_PAINT,
            wx.EVT_SET_FOCUS, wx.EVT_LEFT_DOWN, wx.EVT_LEFT_UP, wx.EVT_MOTION,
            wx.EVT_MOUSEWHEEL, wx.EVT_ENTER_WINDOW, wx.EVT_LEAVE_WINDOW,
            wx.EVT_SIZE )

        if self._text is not None:
            disconnect( self._text, wx.EVT_TEXT_ENTER )
            disconnect_no_id( self._text, wx.EVT_KILL_FOCUS,
                wx.EVT_ENTER_WINDOW, wx.EVT_LEAVE_WINDOW, wx.EVT_CHAR )

        super( _ThemedSliderEditor, self ).dispose()

    #---------------------------------------------------------------------------
    #  Updates the editor when the object trait changes external to the editor:
    #---------------------------------------------------------------------------

    def update_editor ( self ):
        """ Updates the editor when the object trait changes externally to the
            editor.
        """
        self.text       = '%g' % self.value
        self._text_size = None
        self._refresh()

        if self._text is not None:
            self._text.SetValue( self.text )

    #---------------------------------------------------------------------------
    #  Updates the object when the control slider value changes:
    #---------------------------------------------------------------------------

    def update_object ( self, value ):
        """ Updates the object when the control slider value changes.
        """
        try:
            self.value = value
        except TraitError:
            self.value = int( value )

        self.update_editor()

    #---------------------------------------------------------------------------
    #  Handles an error that occurs while setting the object's trait value:
    #---------------------------------------------------------------------------

    def error ( self, excp ):
        """ Handles an error that occurs while setting the object's trait value.
        """
        pass

    #-- Private Methods --------------------------------------------------------

    def _get_text_bounds ( self ):
        """ Get the window bounds of where the current text should be
            displayed.
        """
        tdx, tdy, descent, leading = self._get_text_size()
        wdx, wdy  = self.control.GetClientSizeTuple()
        ty        = ((wdy - (tdy - descent)) / 2) - 2
        alignment = self.factory.alignment
        if alignment == 'left':
            tx = 4
        elif alignment == 'center':
            tx = (wdx - tdx) / 2
        else:
            tx = wdx - tdx - 4

        return ( tx, ty, tdx, tdy )

    def _get_text_size ( self ):
        """ Returns the text size information for the window.
        """
        if self._text_size is None:
            self._text_size = self.control.GetFullTextExtent(
                                               self.text.strip() or 'M' )

        return self._text_size

    def _refresh ( self ):
        """ Refreshes the contents of the control.
        """
        if self.control is not None:
            self.control.Refresh()

    def _set_slider_position ( self, x ):
        """ Calculates a new slider value for a specified (x,y) coordinate.
        """
        wdx, wdy = self.control.GetSizeTuple()
        if 3 <= x < wdx:
            value = self.low + (((x - 3) * (self.high - self.low)) / (wdx - 4))
            increment = self.increment
            if increment > 0:
                value = round( value / increment ) * increment
            self.update_object( value )

    def _delayed_click ( self ):
        """ Handle a delayed click response.
        """
        if self._pending:
            self._pending = False
            self._set_slider_position( self._x )

    def _pop_up_text ( self ):
        """ Pop-up a text control to allow the user to enter a value using
            the keyboard.
        """
        control    = self.control
        self._text = text = wx.TextCtrl( control, -1, self.text,
                            size  = control.GetSize(),
                            style = self.text_styles[ self.factory.alignment ] |
                                    wx.TE_PROCESS_ENTER )
        text.SetSelection( -1, -1 )
        text.SetFocus()
        wx.EVT_TEXT_ENTER( control, text.GetId(), self._text_completed )
        wx.EVT_KILL_FOCUS( text, self._text_completed )
        wx.EVT_ENTER_WINDOW( text, self._enter_text )
        wx.EVT_LEAVE_WINDOW( text, self._leave_text )
        wx.EVT_CHAR( text, self._key_entered )

    def _destroy_text ( self ):
        """ Destroys the current text control.
        """
        self._ignore_focus = self._in_text_window
        disconnect( self._text, wx.EVT_TEXT_ENTER )
        disconnect_no_id( self._text, wx.EVT_KILL_FOCUS, wx.EVT_ENTER_WINDOW,
            wx.EVT_LEAVE_WINDOW, wx.EVT_CHAR )
        self.control.DestroyChildren()
        self._text = None

    #--- wxPython Event Handlers -----------------------------------------------

    def _erase_background ( self, event ):
        """ Do not erase the background here (do it in the 'on_paint' handler).
        """
        pass

    def _on_paint ( self, event ):
        """ Paint the background using the associated ImageSlice object.
        """
        control = self.control
        dc      = BufferDC( control )

        # Draw the slider bar:
        wdx, wdy = control.GetClientSize()
        dx       = max( 0, min( wdx - 2,
                        int( round( ((wdx - 3) * (self.value - self.low)) /
                                                 (self.high - self.low) ) ) ) )

        factory = self.factory
        dc.SetBrush( wx.Brush( factory.slider_color_ ) )
        dc.SetPen( wx.TRANSPARENT_PEN )
        dc.DrawRectangle( 0, 0, dx + 3, wdy )

        # Draw the rest of the background:
        dc.SetBrush( wx.Brush( factory.bg_color_ ) )
        dc.DrawRectangle( dx + 3, 0, wdx - dx - 3, wdy )

        # Draw the slider tip:
        dc.SetBrush( wx.Brush( factory.tip_color_ ) )
        dc.DrawRectangle( dx, 0, 3, wdy )

        # Draw the current text value (if requested):
        if factory.show_value:
            dc.SetBackgroundMode( wx.TRANSPARENT )
            dc.SetTextForeground( factory.text_color_ )
            dc.SetFont( control.GetFont() )
            tx, ty, tdx, tdy = self._get_text_bounds()
            dc.DrawText( self.text, tx, ty )

        # Copy the buffer to the display:
        dc.copy()

    def _resize ( self, event ):
        """ Handles the control being resized.
        """
        if self._text is not None:
            self._text.SetSize( self.control.GetSize() )

    def _set_focus ( self, event ):
        """ Handle the control getting the keyboard focus.
        """
        if ((not self._ignore_focus) and
            (self._x is None)        and
            (self._text is None)):
            self._pop_up_text()

        event.Skip()

    def _left_down ( self, event ):
        """ Handles the left mouse being pressed.
        """
        self._x, self._y = event.GetX(), event.GetY()
        self._pending    = True
        self.control.CaptureMouse()
        do_after( 150, self._delayed_click )

    def _left_up ( self, event ):
        """ Handles the left mouse button being released.
        """
        if self._x is not None:
            self.control.ReleaseMouse()

        if self._pending:
            self._pop_up_text()

        self._x = self._y = self._pending = None

    def _motion ( self, event ):
        """ Handles the mouse moving.
        """
        if self._x is not None:
            x, y = event.GetX(), event.GetY()
            if self._pending:
                if (abs( x - self._x ) + abs( y - self._y )) < 3:
                    return
                self._pending = False
            self._set_slider_position( x )

    def _mouse_wheel ( self, event ):
        """ Handles the mouse wheel rotating.
        """
        if self._in_window:
            increment = event.GetWheelRotation() / event.GetWheelDelta()
            delta     = (self.high - self.low) / 100.0
            if isinstance( self.value, int ) and (abs( delta ) < 1):
                delta = int( abs( delta ) / delta )

            self.update_object( min( max( self.value + increment * delta,
                                          self.low ), self.high ) )

    def _enter_window ( self, event ):
        """ Handles the mouse pointer entering the control.
        """
        self._in_window = True

        if not self._ignore_focus:
            self._ignore_focus = True
            self.control.SetFocus()

        self._ignore_focus = False

    def _leave_window ( self, event ):
        """ Handles the mouse pointer leaving the control.
        """
        self._in_window = False

    def _update_value ( self, event ):
        """ Updates the object value from the current text control value.
        """
        control = event.GetEventObject()
        try:
            self.update_object( float( control.GetValue() ) )

            return True

        except TraitError:
            control.SetBackgroundColour( ErrorColor )
            control.Refresh()

            return False

    def _enter_text ( self, event ):
        """ Handles the mouse entering the pop-up text control.
        """
        self._in_text_window = True

    def _leave_text ( self, event ):
        """ Handles the mouse leaving the pop-up text control.
        """
        self._in_text_window = False

    def _text_completed ( self, event ):
        """ Handles the user pressing the 'Enter' key in the text control.
        """
        if self._update_value( event ):
            self._destroy_text()

    def _key_entered ( self, event ):
        """ Handles individual key strokes while the text control is active.
        """
        key_code = event.GetKeyCode()
        if key_code == wx.WXK_ESCAPE:
            self._destroy_text()
            return

        if key_code == wx.WXK_TAB:
            if self._update_value( event ):
                if event.ShiftDown():
                    self.control.Navigate( 0 )
                else:
                    self.control.Navigate()
            return

        event.Skip()

#-------------------------------------------------------------------------------
#  Create the editor factory object:
#-------------------------------------------------------------------------------

# wxPython editor factory for themed slider editors:
class ThemedSliderEditor ( BasicEditorFactory ):

    # The editor class to be created:
    klass = _ThemedSliderEditor

    # The low end of the slider range:
    low = Float

    # The high end of the slider range:
    high = Float

    # The smallest allowed increment:
    increment = Float

    # Should the current value be displayed as text?
    show_value = Bool( True )

    # The alignment of the text within the slider:
    alignment = Alignment( 'center' )

    # The color to use for the slider bar:
    slider_color = Color( 0xC0C0C0 )

    # The background color for the slider:
    bg_color = Color( 'white' )

    # The color of the slider tip:
    tip_color = Color( 0xFF7300 )

    # The color to use for the value text:
    text_color = Color( 'black' )




#------------------------------------------------------------------------------
#
#  Copyright (c) 2008, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#
#------------------------------------------------------------------------------
""" Defines the tuple editor factory for all traits user interface toolkits.
"""

#-------------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------------

from __future__ import absolute_import

from traits.trait_base import SequenceTypes

from traits.api import Bool, HasTraits, List, Tuple, Unicode, Int, Any, TraitType

# CIRCULAR IMPORT FIXME: Importing from the source rather than traits.ui.api
# to avoid circular imports, as this EditorFactory will be part of
# traits.ui.api as well.
from ..view import View

from ..group import Group

from ..item import Item

from ..editor_factory import EditorFactory

from ..editor import Editor

#-------------------------------------------------------------------------------
#  'ToolkitEditorFactory' class:
#-------------------------------------------------------------------------------

class ToolkitEditorFactory ( EditorFactory ):
    """ Editor factory for tuple editors.
    """
    #---------------------------------------------------------------------------
    #  Trait definitions:
    #---------------------------------------------------------------------------

    # Trait definitions for each tuple field
    types = Any

    # Labels for each of the tuple fields
    labels = List( Unicode )

    # Editors for each of the tuple fields:
    editors = List( EditorFactory )

    # Number of tuple fields or rows
    cols   = Int( 1 )

    # Is user input set on every keystroke? This is applied to every field
    # of the tuple, provided the field does not already have an 'auto_set'
    # metadata or an editor defined.
    auto_set = Bool( True )

    # Is user input set when the Enter key is pressed? This is applied to
    # every field of the tuple, provided the field does not already have an
    # 'enter_set' metadata or an editor defined.
    enter_set = Bool( False )

#-------------------------------------------------------------------------------
#  'SimpleEditor' class:
#-------------------------------------------------------------------------------

class SimpleEditor ( Editor ):
    """ Simple style of editor for tuples.

    The editor displays an editor for each of the fields in the tuple, based on
    the type of each field.
    """
    #---------------------------------------------------------------------------
    #  Finishes initializing the editor by creating the underlying toolkit
    #  widget:
    #---------------------------------------------------------------------------

    def init ( self, parent ):
        """ Finishes initializing the editor by creating the underlying toolkit
            widget.
        """
        self._ts     = ts = TupleStructure( self )
        self._ui     = ui = ts.view.ui( ts, parent, kind = 'subpanel' ).set(
                                        parent = self.ui )
        self.control = ui.control
        self.set_tooltip()

    #---------------------------------------------------------------------------
    #  Updates the editor when the object trait changes external to the editor:
    #---------------------------------------------------------------------------

    def update_editor ( self ):
        """ Updates the editor when the object trait changes external to the
            editor.
        """
        ts = self._ts
        for i, value in enumerate( self.value ):
            setattr( ts, 'f%d' % i, value )

    #---------------------------------------------------------------------------
    #  Returns the editor's control for indicating error status:
    #---------------------------------------------------------------------------

    def get_error_control ( self ):
        """ Returns the editor's control for indicating error status.
        """
        return self._ui.get_error_controls()

#-------------------------------------------------------------------------------
#  'TupleStructure' class:
#-------------------------------------------------------------------------------

class TupleStructure ( HasTraits ):
    """ Creates a view containing items for each field in a tuple.
    """
    #---------------------------------------------------------------------------
    #  Trait definitions:
    #---------------------------------------------------------------------------

    # Editor this structure is linked to
    editor = Any

    # The constructed View for the tuple
    view = Any

    # Number of tuple fields
    fields = Int

    #---------------------------------------------------------------------------
    #  Initializes the object:
    #---------------------------------------------------------------------------

    def __init__ ( self, editor ):
        """ Initializes the object.
        """
        factory = editor.factory
        types   = factory.types
        labels  = factory.labels
        editors = factory.editors
        cols    = factory.cols

        # Save the reference to the editor:
        self.editor = editor

        # Get the tuple we are mirroring:
        object = editor.value

        # For each tuple field, add a trait with the appropriate trait
        # definition and default value:
        content     = []
        self.fields = len( object )
        len_labels  = len( labels )
        len_editors = len( editors )

        if types is None:
            type = editor.value_trait.handler
            if isinstance( type, Tuple ):
                types = type.types

        if not isinstance( types, SequenceTypes ):
            types = [ types ]

        len_types = len( types )
        if len_types == 0:
            types     = [ Any ]
            len_types = 1

        for i, value in enumerate( object ):
            type = types[ i % len_types ]

            auto_set = enter_set = None
            if isinstance(type, TraitType):
                auto_set = type.auto_set
                enter_set = type.enter_set
            if auto_set is None:
                auto_set = editor.factory.auto_set
            if enter_set is None:
                enter_set = editor.factory.enter_set

            label = ''
            if i < len_labels:
                label = labels[i]

            field_editor = None
            if i < len_editors:
                field_editor = editors[i]

            name = 'f%d' % i
            self.add_trait( name, type( value, event = 'field',
                                               auto_set = auto_set,
                                               enter_set = enter_set ) )
            item = Item( name = name, label = label, editor = field_editor )
            if cols <= 1:
                content.append( item )
            else:
                if (i % cols) == 0:
                    group = Group( orientation = 'horizontal' )
                    content.append( group )

                group.content.append( item )

        self.view = View( Group( show_labels = (len_labels != 0), *content ) )

    #---------------------------------------------------------------------------
    #  Updates the underlying tuple when any field changes value:
    #---------------------------------------------------------------------------

    def _field_changed ( self, name, old, new ):
        """ Updates the underlying tuple when any field changes value.
        """
        index = int( name[ 1: ] )
        value = self.editor.value
        if new != value[ index ]:
            self.editor.value = tuple( 
                [ getattr( self, 'f%d' % i ) for i in range( self.fields ) ] )


# Define the TupleEditor class.
TupleEditor = ToolkitEditorFactory

### EOF #######################################################################






import os

import jsonpickle

import omnivore.arch.machine as machine


class TestJsonPickle(object):
    def setup(self):
        self.machine = machine.Atari5200

    def test_simple(self):
        data = self.machine.__getstate__()
        print self.machine.all_trait_names()
        t = self.machine.trait('disassembler')
        print t
        print dir(t)
        for name in self.machine.all_trait_names():
            t = self.machine.trait(name)
            try:
                print name, t.transient, getattr(self.machine, name)
            except AttributeError:
                pass
        print data
        encoded = jsonpickle.encode(self.machine)
        print encoded
        decoded = jsonpickle.decode(encoded)
        print decoded
        print "Attributes!"
        for name in self.machine.all_trait_names():
            print name
        for name in decoded.all_trait_names():
            # a trait named "default_factory" shows up somehow in
            # all_trait_names, but not in the object itself, so ignore
            # anything not in the object's dict
            if hasattr(decoded, name):
                print name, getattr(decoded, name)
        print
        assert self.machine.disassembler == decoded.disassembler
        print self.machine.memory_map
        print decoded.memory_map
        assert self.machine.memory_map == decoded.memory_map
        assert self.machine == decoded

if __name__ == "__main__":
    t = TestJsonPickle()
    t.setup()
    t.test_simple()



from omnivore.framework.panes import FrameworkPane

# Local imports.
from omnivore.tasks.hex_edit.segments import SegmentList
from omnivore.tasks.hex_edit.grid_control import HexEditControl
from omnivore.tasks.hex_edit.panes import CommentsPanel
from omnivore.framework.undo_panel import UndoHistoryPanel
from omnivore.utils.wx.springtabs import SpringTabs
from omnivore.utils.wx.info_panels import InfoPanel

from peanuts import TriggerList

import logging
log = logging.getLogger(__name__)


class SegmentsPane(FrameworkPane):
    #### TaskPane interface ###################################################

    id = 'jumpman.segments'
    name = 'Segments'
    
    def create_contents(self, parent):
        control = SegmentList(parent, self.task, size=(64,150))
        return control


class UndoPane(FrameworkPane):
    #### TaskPane interface ###################################################

    id = 'jumpman.undo'
    name = 'Undo History'
    
    def create_contents(self, parent):
        control = UndoHistoryPanel(parent, self.task, size=(64,150))
        return control


class HexPane(FrameworkPane):
    #### TaskPane interface ###################################################

    id = 'jumpman.hex'
    name = 'Raw Level Data'
    
    def create_contents(self, parent):
        control = HexEditControl(parent, self.task, size=(350, 150))
        return control


class TriggerPane(FrameworkPane):
    #### TaskPane interface ###################################################

    id = 'jumpman.triggers'
    name = 'Trigger Painting'
    
    def create_contents(self, parent):
        control = TriggerList(parent, self.task, size=(350,150))
        return control


class JumpmanInfoPanel(InfoPanel):
    def is_valid_data(self):
        return self.editor.valid_jumpman_segment and bool(self.editor.bitmap.level_builder.objects)


class LevelDataPane(FrameworkPane):
    #### TaskPane interface ###################################################

    id = 'jumpman.level_data'
    name = 'Level Data'
    
    def create_contents(self, parent):
        fields = [
            ("text", "Level Number", 0x00, 2),
            ("atascii_gr2_0xc0", "Level Name", 0x3ec, 20),
            ("uint", "Points per Peanut", 0x33, 2, 250),
            ("label", "# Peanuts", "num_peanuts", 1000),
            ("peanuts_needed", "Peanuts Needed", 0x3e, 1, ["All", "All except 1", "All except 2", "All except 3", "All except 4"]),
            ("uint", "Bonus Value", 0x35, 2, 2500),
            ("dropdown", "Number of Bullets", 0x3d, 1, ["None", "1", "2", "3", "4"]),
            ("antic_colors", "Game Colors", 0x2a, 9),
            ("label", "# Columns with Ladders", "num_ladders", 12),
            ("label", "# Columns with Downropes", "num_downropes", 6),
        ]
        control = JumpmanInfoPanel(parent, self.task, fields, size=(350, 150))
        return control


class SidebarPane(FrameworkPane):
    #### TaskPane interface ###################################################

    id = 'jumpman.sidebar'
    name = 'Sidebar'
    
    movable = False
    caption_visible = False
    dock_layer = 9
    
    def comments_cb(self, parent, task, **kwargs):
        control = CommentsPanel(parent, task)
        
    def create_contents(self, parent):
        control = SpringTabs(parent, self.task, popup_direction="left")
        control.addTab("Comments", self.comments_cb)
        return control
    
    def refresh_active(self):
        active = self.control._radio
        if active is not None and active.is_shown:
            active.managed_window.refresh_view()



#------------------------------------------------------------------------------
#
#  Copyright (c) 2008, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   10/21/2004
#
#------------------------------------------------------------------------------
""" Defines the table editor factory for all traits user interface toolkits.
"""

#-------------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------------

from __future__ import absolute_import

from traits.api import (Int, Float, List, Instance, Str, Color, Font, Any, Tuple,
        Dict, Enum, Trait, Bool, Callable, Range, on_trait_change)

from ..editor_factory import EditorFactory

from ..handler import Handler

from ..helper import Orientation

from ..item import Item

from ..table_filter import TableFilter

from ..ui_traits import AView

from ..view import View

from .enum_editor import EnumEditor

#-------------------------------------------------------------------------------
#  Constants:
#-------------------------------------------------------------------------------

# The filter used to indicate that the user wants to customize the current
# filter
customize_filter = TableFilter( name = 'Customize...' )

#-------------------------------------------------------------------------------
#  Trait definitions:
#-------------------------------------------------------------------------------

# A trait whose value can be True, False, or a callable function
BoolOrCallable = Trait( False, Bool, Callable )

#-------------------------------------------------------------------------------
#  'ToolkitEditorFactory' class:
#-------------------------------------------------------------------------------

class ToolkitEditorFactory ( EditorFactory ):
    """ Editor factory for table editors.
    """

    #---------------------------------------------------------------------------
    #  Trait definitions:
    #---------------------------------------------------------------------------

    # List of initial table column descriptors
    columns = List( Instance('traitsui.table_column.TableColumn') )

    # List of other table column descriptors (not initially displayed)
    other_columns = List(
                     Instance('traitsui.table_column.TableColumn') )

    # The object trait containing the list of column descriptors
    columns_name = Str

    # The desired number of visible rows in the table
    rows = Int

    # The optional extended name of the trait used to specify an external filter
    # for the table data. The value of the trait must either be an instance of
    # TableEditor, a callable that accepts one argument (a table row) and
    # returns True or False to indicate whether the specified object passes the
    # filter or not, or **None** to indicate that no filter is to be applied:
    filter_name = Str

    # Initial filter that should be applied to the table
    filter = Instance( 'traitsui.table_filter.TableFilter' )

    # List of available filters that can be applied to the table
    filters = List( Instance(
                     'traitsui.table_filter.TableFilter' ) )

    # The optional extended trait name of the trait used to notify that the
    # filter has changed and the displayed objects should be updated.
    # It should be an Event.
    update_filter_name = Str

    # Filter object used to allow a user to search the table.
    # NOTE: If left as None, the table will not be searchable.
    search = Instance( 'traitsui.table_filter.TableFilter' )

    # Default context menu to display when any cell is right-clicked
    menu = Instance( 'traitsui.menu.Menu' )

    # Default trait name containg menu
    menu_name = Str

    # Are objects deletable from the table?
    deletable = BoolOrCallable( False )

    # Is the table editable?
    editable = Bool( True )

    # Should the editor become active after the first click
    edit_on_first_click = Bool( True )

    # Can the user reorder the items in the table?
    reorderable = Bool( False )

    # Can the user configure the table columns?
    configurable = Bool( True )

    # Should the cells of the table automatically size to the optimal size?
    auto_size = Bool( True )

    # Mirrors the Qt QSizePolicy.Policy attribute, for horizontal and vertical
    # dimensions.  For these to be useful, set auto_size to False.  If these
    # are None, then the table size policy will not be set in that dimension
    # (for backwards compatibility).
    h_size_policy = Enum( None, "preferred", "fixed", "minimum", "maximum", "expanding",
                        "minimum_expanding", "ignored" )
    v_size_policy = Enum( None, "preferred", "fixed", "minimum", "maximum", "expanding",
                        "minimum_expanding", "ignored" )

    # Should a new row automatically be added to the end of the table to allow
    # the user to create new entries? If True, **row_factory** must be set.
    auto_add = Bool( False )

    # Should the table items be presented in reverse order?
    reverse = Bool( False )

    # The DockWindow graphical theme:
    dock_theme = Any

    # View to use when editing table items.
    # NOTE: If not specified, the table items are not editable in a separate
    # pane of the editor.
    edit_view = AView( ' ' )

    # The handler to apply to **edit_view**
    edit_view_handler = Instance( Handler )

    # Width to use for the edit view
    edit_view_width = Float( -1.0 )

    # Height to use for the edit view
    edit_view_height = Float( -1.0 )

    # Layout orientation of the table and its associated editor pane. This
    # attribute applies only if **edit_view** is not ' '.
    orientation = Orientation

    # Is the table sortable by clicking on the column headers?
    sortable = Bool( True )

    # Does sorting affect the model (vs. just the view)?
    sort_model = Bool( False )

    # Should grid lines be shown on the table?
    show_lines = Bool( True )

    # Should the toolbar be displayed? (Note that False will override settings
    # such as 'configurable', etc., and is a quick way to prevent the toolbar
    # from being displayed; but True will not cause a toolbar to appear if one
    # would not otherwise have been displayed)
    show_toolbar = Bool( False )

    # The vertical scroll increment for the table:
    scroll_dy = Range( 1, 32 )

    # Grid line color
    line_color = Color( 0xC4C0A9 )

    # Show column labels?
    show_column_labels = Bool( True )

    # Show row labels?
    show_row_labels = Bool( False )

    # Font to use for text in cells
    cell_font = Font

    # Color to use for text in cells
    cell_color = Color( 'black' )

    # Color to use for cell backgrounds
    # The default is the "WindowColor" constant declared in ui.api:
    # we shall set the value in a trait initializer method, in order to avoid
    # circular imports.
    cell_bg_color = Color

    # Color to use for read-only cell backgrounds
    cell_read_only_bg_color = Color( 0xF8F7F1 )

    # Whether to even-odd alternate the background color.
    alternate_bg_color = Bool(False)

    # Font to use for text in labels
    label_font = Font

    # Color to use for text in labels
    label_color = Color( 'black' )

    # Color to use for label backgrounds
    # The default is the "WindowColor" constant declared in ui.api:
    # we shall set the value in a trait initializer method, in order to avoid
    # circular imports.
    label_bg_color = Color

    # Background color of selected item
    selection_bg_color = Color( 'light blue', allow_none = True )

    # Color of selected text
    selection_color = Color( 'black' )

    # Height (in pixels) of column labels
    column_label_height = Int( 25 )

    # Width (in pixels) of row labels
    row_label_width = Int( 82 )

    # The initial height of each row (<= 0 means use default value):
    row_height = Int( 0 )

    # The optional extended name of the trait that the indices of the items
    # currently passing the table filter are synced with:
    filtered_indices = Str

    # The selection mode of the table. The meaning of the various values are as
    # follows:
    #
    # row
    #   Entire rows are selected. At most one row can be selected at once.
    #   This is the default.
    # rows
    #   Entire rows are selected. More than one row can be selected at once.
    # column
    #   Entire columns are selected. At most one column can be selected at
    #   once.
    # columns
    #   Entire columns are selected. More than one column can be selected at
    #   once.
    # cell
    #   Single cells are selected. Only one cell can be selected at once.
    # cells
    #   Single cells are selected. More than one cell can be selected at once.
    selection_mode = Enum( 'row', 'rows', 'column', 'columns', 'cell', 'cells' )

    # The optional extended name of the trait that the current selection is
    # synced with:
    selected = Str

    # The optional extended trait name of the trait that the indices of the
    # current selection are synced with:
    selected_indices = Str

    # The optional extended trait name of the trait that should be assigned
    # an ( object, column ) tuple when a table cell is clicked on (Note: If you
    # want to receive repeated clicks on the same cell, make sure the trait is
    # defined as an Event):
    click = Str

    # The optional extended trait name of the trait that should be assigned
    # an ( object, column ) tuple when a table cell is double-clicked on
    # (Note: if you want to receive repeated double-clicks on the same cell,
    # make sure the trait is defined as an Event):
    dclick = Str

    # Called when a table item is selected
    on_select = Any

    # Called when a table item is double clicked
    on_dclick = Any

    # A factory to generate new rows.
    # NOTE: If None, then the user will not be able to add new rows to the
    # table. If not None, then it must be a callable that accepts
    # **row_factory_args** and **row_factory_kw** and returns a new object
    # that can be added to the table.
    row_factory = Any

    # Arguments to pass to the **row_factory** callable when a new row is
    # created
    row_factory_args = Tuple

    # Keyword arguments to pass to the **row_factory** callable when a new row
    # is created
    row_factory_kw = Dict

    # Hooks for replacing parts of the implementation.
    table_view_factory = Callable()
    source_model_factory = Callable()
    model_factory = Callable()

    #---------------------------------------------------------------------------
    #  Traits view definitions:
    #---------------------------------------------------------------------------

    traits_view = View( [ '{Initial columns}@',
                          Item( 'columns',       resizable = True ),
                          '{Other columns}@',
                          Item( 'other_columns', resizable = True ),
                          '|{Columns}<>' ],
                        [ [ 'deletable{Are items deletable?}', '9',
                            'editable{Are items editable?}',   '9',
                            '-[Item Options]>' ],
                          [ 'show_column_labels{Show column labels?}',      '9',
                            'configurable{Are columns user configurable?}', '9',
                            'auto_size{Should columns auto size?}',
                            '-[Column Options]>' ],
                          [ 'sortable{Are columns sortable?}',
                            Item( 'sort_model{Does sorting affect the model?}',
                                  enabled_when = 'sortable' ),
                            '-[Sorting Options]>' ],
                          [ [ 'show_lines{Show grid lines?}',
                              '|>' ],
                            [ '_', 'line_color{Grid line color}@',
                              '|<>' ],
                            '|[Grid Line Options]' ],
                          '|{Options}' ],
                        [ [ 'cell_color{Text color}@',
                            'cell_bg_color{Background color}@',
                            'cell_read_only_bg_color{Read only color}@',
                            '|[Cell Colors]' ],
                          [ 'cell_font',
                            '|[Cell Font]<>' ],
                          '|{Cell}' ],
                        [ [ 'label_color{Text color}@',
                            'label_bg_color{Background color}@',
                            '|[Label Colors]' ],
                          [ 'label_font@',
                            '|[Label Font]<>' ],
                          '|{Label}' ],
                        [ [ 'selection_color{Text color}@',
                            'selection_bg_color{Background color}@',
                            '|[Selection Colors]' ],
                          '|{Selection}' ],
                        height = 0.5 )

    #---------------------------------------------------------------------------
    #  'Editor' factory methods:
    #---------------------------------------------------------------------------

    def readonly_editor ( self, ui, object, name, description, parent ):
        """ Generates an "editor" that is read-only.
        Overridden to set the value of the editable trait to False before
        generating the editor.

        """
        self.editable = False
        return super(ToolkitEditorFactory, self).readonly_editor(
                  ui, object, name, description, parent)

    def _cell_bg_color_default(self):
        """ Returns the default value of the cell background color.
        """
        # NOTE: We are initializing the 'cell_bg_color' trait in this method
        # instead of in the trait definition so as to delay importing from
        # ui.api until needed (will lead to circular imports otherwise).
        from ..api import WindowColor
        return WindowColor

    def _label_bg_color_default(self):
        """ Returns the default value of the cell background color.
        """
        # NOTE: We are initializing the 'cell_bg_color' trait in this method
        # instead of in the trait definition so as to delay importing from
        # ui.api until needed (will lead to circular imports otherwise).
        from ..api import WindowColor
        return WindowColor

    #---------------------------------------------------------------------------
    #  Event handlers:
    #---------------------------------------------------------------------------

    @on_trait_change('filters[]')
    def _update_filter_editor ( self, object, name, old, new ):
        """ Handles the set of filters associated with the editor's factory
            being changed.
        """
        values = { None: '000:No filter' }
        i      = 0
        for filter in self.filters:
            if not filter.template:
                i += 1
                values[ filter ] = '%03d:%s' % ( i, filter.name )
        values[ customize_filter ] = '%03d:%s' % ( (i + 1),
                                                   customize_filter.name )
        if self._filter_editor is None:
            self._filter_editor = EnumEditor( values = values )
        else:
            self._filter_editor.values = values


# Define the TableEditor class
TableEditor = ToolkitEditorFactory

#-------------------------------------------------------------------------------
#  Base class for toolkit-specific editors
#-------------------------------------------------------------------------------

class BaseTableEditor(object):
    """ Base class for toolkit-specific editors.
    """

    #---------------------------------------------------------------------------
    #  Interface for toolkit-specific editors:
    #---------------------------------------------------------------------------

    def set_menu_context ( self, selection, object, column ):
        """Call before creating a context menu for a cell, then set self as the
           controller for the menu.
        """
        self._menu_context = { 'selection': selection,
                               'object':    object,
                               'column':    column,
                               'editor':    self,
                               'info':      self.ui.info,
                               'handler':   self.ui.handler }

    #---------------------------------------------------------------------------
    #  pyface.action 'controller' interface implementation:
    #---------------------------------------------------------------------------

    def add_to_menu ( self, menu_item ):
        """ Adds a menu item to the menu bar being constructed.
        """
        action = menu_item.item.action
        self.eval_when( action.enabled_when, menu_item, 'enabled' )
        self.eval_when( action.checked_when, menu_item, 'checked' )

    def add_to_toolbar ( self, toolbar_item ):
        """ Adds a toolbar item to the toolbar being constructed.
        """
        self.add_to_menu( toolbar_item )

    def can_add_to_menu ( self, action ):
        """ Returns whether the action should be defined in the user interface.
        """
        if action.defined_when != '':
            if not eval( action.defined_when, globals(),
                         self._menu_context ):
                return False

        if action.visible_when != '':
            if not eval( action.visible_when, globals(),
                         self._menu_context ):
                return False

        return True

    def can_add_to_toolbar ( self, action ):
        """ Returns whether the toolbar action should be defined in the user
            interface.
        """
        return self.can_add_to_menu( action )

    def perform ( self, action, action_event = None ):
        """ Performs the action described by a specified Action object.
        """
        self.ui.do_undoable( self._perform, action )

    def _perform ( self, action ):
        method_name        = action.action
        info               = self.ui.info
        handler            = self.ui.handler
        context            = self._menu_context
        self._menu_context  = None
        selection          = context[ 'selection' ]

        if method_name.find( '.' ) >= 0:
            if method_name.find( '(' ) < 0:
                method_name += '()'
            try:
                eval( method_name, globals(), context )
            except:
                # fixme: Should the exception be logged somewhere?
                pass
            return

        method = getattr( handler, method_name, None )
        if method is not None:
            method( info, selection )
            return

        if action.on_perform is not None:
            action.on_perform( selection )
            return

        action.perform( selection )

    #---------------------------------------------------------------------------
    #  Menu support methods:
    #---------------------------------------------------------------------------

    def eval_when ( self, condition, object, trait ):
        """ Evaluates a condition within a defined context and sets a specified
        object trait based on the result, which is assumed to be a Boolean.
        """
        if condition != '':
            value = bool( eval( condition, globals(), self._menu_context ) )
            setattr( object, trait, value )

#-------------------------------------------------------------------------------
#  Helper class for toolkit-specific editors to implement 'reversed' option:
#-------------------------------------------------------------------------------

class ReversedList ( object ):
    """ A list whose order is the reverse of its input.
    """

    def __init__ ( self, list ):
        self.list = list

    def insert ( self, index, value ):
        """ Inserts a value at a specified index in the list.
        """
        return self.list.insert( self._index( index - 1 ), value )

    def index ( self, value ):
        """ Returns the index of the first occurrence of the specified value in
            the list.
        """
        list = self.list[:]
        list.reverse()

        return list.index( value )

    def __len__ ( self ):
        """ Returns the length of the list.
        """
        return len( self.list )

    def __getitem__ ( self, index ):
        """ Returns the value at a specified index in the list.
        """
        return self.list[ self._index( index ) ]

    def __setslice__ ( self, i, j, values ):
        """ Sets a slice of a list to the contents of a specified sequence.
        """
        return self.list.__setslice__( self._index( i ), self._index( j ),
                                       values )

    def __delitem__ ( self, index ):
        """ Deletes the item at a specified index.
        """
        return self.list.__delitem__( self._index( index ) )

    def _index ( self, index ):
        """ Returns the "reversed" value for a specified index.
        """
        if index < 0:
            return (-1 - index)

        result = (len( self.list ) - index - 1)
        if result >= 0:
            return result

        return index



import os
import sys
import time
import argparse
import json
import jsonpickle
from datetime import datetime

import fs

from filesystem import init_filesystems
from document import Document

import logging
log = logging.getLogger(__name__)

# Create the wx app here so we can capture the Mac specific stuff that
# traitsui.wx.toolkit's wx creation routines don't handle.  Also, monkey
# patching wx.GetApp() to add MacOpenFile (etc) doesn't seem to work, so we
# have to have these routines at app creation time.
import wx
class EnthoughtWxApp(wx.App):
    def MacOpenFiles(self, filenames):
        """OSX specific routine to handle files that are dropped on the icon
        
        """
        if hasattr(self, 'tasks_application'):
            # The tasks_application attribute is added to this wx.App instance
            # when the application has been initialized.  This is used as a
            # flag to indicate that the subsequent calls to MacOpenFile are
            # real drops of files onto the dock icon.  Prior to that, this
            # method gets called for all the command line arguments which would
            # give us two copies of each file specified on the command line.
            for filename in filenames:
                log.debug("MacOpenFile: loading %s" % filename)
                self.tasks_application.load_file(filename, None)
        else:
            log.debug("MacOpenFile: skipping %s because it's a command line argument" % filename)
    
    throw_out_next_wheel_rotation = False
    
    def FilterEvent(self, evt):
        if hasattr(evt, "GetWheelRotation"):
            print "FILTEREVENT!!!", evt.GetWheelRotation(), evt
            wheel = evt.GetWheelRotation()
            if wheel != 0:
                if self.throw_out_next_wheel_rotation:
                    self.throw_out_next_wheel_rotation = False
                    return 0
                self.throw_out_next_wheel_rotation = True
        if hasattr(evt, "GetKeyCode"):
            print "FILTEREVENT!!! char=%s, key=%s, modifiers=%s" % (evt.GetUniChar(), evt.GetKeyCode(), bin(evt.GetModifiers()))
        return -1

from traits.etsconfig.api import ETSConfig

_app = EnthoughtWxApp(redirect=False)
if False:  # enable this to use FilterEvent
    _app.SetCallFilterEvent(True)

# Enthought library imports.
from envisage.ui.tasks.api import TasksApplication
from envisage.ui.tasks.task_window_event import TaskWindowEvent, VetoableTaskWindowEvent
from pyface.api import ImageResource
from pyface.tasks.api import Task, TaskWindowLayout
from traits.api import provides, Bool, Instance, List, Property, Str, Unicode, Event, Dict, Int, Float, Tuple

# Local imports.
from omnivore.framework.preferences import FrameworkPreferences, \
    FrameworkPreferencesPane


def _task_window_wx_on_mousewheel(self, event):
    if self.active_task and hasattr(self.active_task, '_wx_on_mousewheel_from_window'):
        log.debug("calling mousewheel in task %s" % self.active_task)
        self.active_task._wx_on_mousewheel_from_window(event)

class FrameworkApplication(TasksApplication):
    """ The sample framework Tasks application.
    """

    #### 'IApplication' interface #############################################

    # The application's globally unique identifier.
    id = 'omnivore.framework.application'

    # The application's user-visible name.
    name = 'Omnivore'

    #### 'TasksApplication' interface #########################################

    # The default window-level layout for the application.
    default_layout = List(TaskWindowLayout)

    # Whether to restore the previous application-level layout when the
    # applicaton is started.
    always_use_default_layout = Property(Bool)

    #### 'FrameworkApplication' interface ####################################

    preferences_helper = Instance(FrameworkPreferences)
    
    startup_task = Str('omnivore.framework.text_edit_task')
    
    successfully_loaded_event = Event
    
    plugin_event = Event
    
    preferences_changed_event = Event
    
    plugin_data = {}
    
    command_line_args = List
    
    log_dir = Str
    
    log_file_ext = Str
    
    cache_dir = Str
    
    next_document_id = Int(0)
    
    documents = List
    
    last_clipboard_check_time = Float(-1)
    
    clipboard_check_interval = Float(.75)
    
    perspectives = Dict
    
    default_window_size = (800, 600)
    
    last_window_size = Tuple(default_window_size)

    ###########################################################################
    # Private interface.
    ###########################################################################

    #### Trait initializers ###################################################
    
    def _about_title_default(self):
        return self.name
    
    def _about_version_default(self):
        from omnivore import __version__
        return __version__

    def _default_layout_default(self):
        active_task = self.preferences_helper.default_task
        log.debug("active task: -->%s<--" % active_task)
        if not active_task:
            active_task = self.startup_task
        log.debug("active task: -->%s<--" % active_task)
        log.debug("factories: %s" % " ".join([ factory.id for factory in self.task_factories]))
        tasks = [ factory.id for factory in self.task_factories if active_task and active_task == factory.id ]
        log.debug("Default layout: %s" % str(tasks))
        return [ TaskWindowLayout(*tasks,
                                  active_task = active_task,
                                  size = self.last_window_size) ]

    def _preferences_helper_default(self):
        return FrameworkPreferences(preferences = self.preferences)

    #### Trait property getter/setters ########################################

    def _get_always_use_default_layout(self):
        #return self.preferences_helper.always_use_default_layout
        return True


    #### Trait event handlers
    
    def _application_initialized_fired(self):
        log.debug("STARTING!!!")
        init_filesystems()
        loaded = False
        parser = argparse.ArgumentParser(description="Application parser")
        parser.add_argument("-t", "--task_id", "--edit_with", action="store", default="", help="Use the editing mode specified by this task id for all files listed on the command line")
        parser.add_argument("--show_editors", action="store_true", default=False, help="List all task ids")
        options, extra_args = parser.parse_known_args(self.command_line_args)
        if options.show_editors:
            for factory in self.task_factories:
                print("%s %s" % (factory.id, factory.name))
        for arg in extra_args:
            if arg.startswith("-"):
                log.debug("skipping flag %s" % arg)
                continue
            log.debug("processing %s" % arg)
            task_id = self.find_best_task_id(options.task_id)
            self.load_file(arg, None, task_id=task_id)
            loaded = True
        if not loaded:
            factory = self.get_task_factory(self.startup_task)
            url = factory.factory.about_application
            if url:
                self.load_file(url)
        app = wx.GetApp()
        app.tasks_application = self
        
        app.Bind(wx.EVT_IDLE, self.on_idle)
    
    def on_idle(self, evt):
        evt.Skip()
        if not self.active_window:
            return
        editor = self.active_window.active_task.active_editor
        if editor is None:
            return
        t = time.time()
        if t > self.last_clipboard_check_time + self.clipboard_check_interval:
            wx.CallAfter(self.check_clipboard_can_paste, editor)
            self.last_clipboard_check_time = time.time()
        editor.perform_idle()
    
    def check_clipboard_can_paste(self, editor):
        data_formats = [o.GetFormat() for o in editor.supported_clipboard_data_objects]
        log.debug("Checking clipboard formats %s" % str(data_formats))
        supported = False
        if not wx.TheClipboard.IsOpened():
            try:
                if wx.TheClipboard.Open():
                    for f in data_formats:
                        if wx.TheClipboard.IsSupported(f):
                            log.debug("  found clipboard format: %s" % str(f))
                            supported = True
            finally:
                if wx.TheClipboard.IsOpened():
                    wx.TheClipboard.Close()
            editor.can_paste = supported
        
    def _window_created_fired(self, event):
        """The toolkit window doesn't exist yet.
        """
        self.init_perspectives()
    
    def _window_opened_fired(self, event):
        """The toolkit window does exist here.
        """
        log.debug("WINDOW OPENED!!! %s, size=%s" % (event.window.control, str(self.last_window_size)))
        event.window.size = tuple(self.last_window_size)
        
        # Check to see that there's at least one task.  If a bad application
        # memento (~/.config/Omnivore/tasks/wx/application_memento), the window
        # may be blank in which case we need to add the default task.
        if not event.window.tasks:
            self.create_task_in_window(self.startup_task, event.window)
            log.debug("EMPTY WINDOW OPENED!!! Created task.")
        
        task = event.window.active_task
        if task.active_editor is None and task.start_new_editor_in_new_window:
            task.new()
        
        if sys.platform.startswith("win"):
            # monkey patch to include mousewheel handler on the TaskWindow
            import types
            event.window._wx_on_mousewheel = types.MethodType(_task_window_wx_on_mousewheel, event.window)
            event.window.control.Bind(wx.EVT_MOUSEWHEEL, event.window._wx_on_mousewheel)

    #### API

    def load_file(self, uri, active_task=None, task_id="", in_current_window=False, **kwargs):
        service = self.get_service("omnivore.file_type.i_file_recognizer.IFileRecognizerDriver")
        log.debug("SERVICE!!! %s" % service)
        
        from omnivore.utils.file_guess import FileGuess
        # The FileGuess loads the first part of the file and tries to identify it.
        try:
            guess = FileGuess(uri)
        except fs.errors.FSError, e:
            log.error("File load error: %s" % str(e))
            if active_task is not None:
                active_task.window.error(str(e), "File Load Error")
            return
        
        # Attempt to classify the guess using the file recognizer service
        document = service.recognize(guess)
        log.debug("created document %s (mime=%s) %d segments from parser %s" % (document, document.metadata.mime, len(document.segments), document.segment_parser.__class__.__name__))
        
        # Short circuit: if the file can be edited by the active task, use that!
        if active_task is not None and active_task.can_edit(document):
            log.debug("active task %s can edit %s" % (active_task, document))
            active_task.new(document, **kwargs)
            return
        
        possibilities = self.get_possible_task_factories(document, task_id)
        if not possibilities:
            log.debug("no editor for %s" % uri)
            return
        best = self.find_best_task_factory(document, possibilities)
        
        if active_task is not None:
            # Ask the active task if it's OK to load a different editor
            if not active_task.allow_different_task(guess, best.factory):
                return
            dummy = Document(metadata="application/octet-stream")
            if active_task.can_edit(document) and active_task.ask_attempt_loading_as_octet_stream(guess, best.factory):
                log.debug("Active task %s allows application/octet-stream" % active_task.id)
                active_task.new(document, **kwargs)
                return
            if in_current_window:
                task = self.create_task_in_window(best.id, active_task.window)
                task.new(document, **kwargs)
                return

        # Look for existing task in current windows
        task = self.find_active_task_of_type(best.id)
        if task:
            log.debug("Found task %s in current window" % best.id)
            task.new(document, **kwargs)
            return
        
        log.debug("Creating task %s in current window" % best.id)
        self.create_task_from_factory_id(document, best.id)
    
    def get_possible_task_factories(self, document, task_id=""):
        possibilities = []
        for factory in self.task_factories:
            log.debug("factory: %s=%s" % (factory.id, factory.name))
            if task_id:
                if factory.id == task_id:
                    possibilities.append(factory)
            elif hasattr(factory.factory, "can_edit"):
                if factory.factory.can_edit(document):
                    log.debug("  can edit: %s" % document)
                    possibilities.append(factory)
        log.debug("get_possible_task_factories: %s" % str([(p.name, p.id) for p in possibilities]))
        return possibilities
    
    def find_best_task_factory(self, document, factories):
        scores = []
        for factory in factories:
            log.debug("factory: %s=%s" % (factory.id, factory.name))
            if document.last_task_id == factory.id:
                # short circuit if document is requesting a specific task
                return factory
            score = factory.factory.get_match_score(document)
            scores.append((score, factory))
        scores.sort()
        log.debug("find_best_task_factory: %s" % str([(s, p.name, p.id) for (s, p) in scores]))
        return scores[-1][1]
    
    def get_task_factory(self, task_id):
        for factory in self.task_factories:
            if factory.id == task_id:
                return factory
        return None
    
    def find_best_task_id(self, task_id):
        if task_id:
            for factory in self.task_factories:
                if factory.id == task_id or ".%s." % task_id in factory.id or ".%s" % task_id in factory.id:
                    return factory.id
        return ""  # empty string will result in scanning the file for the best match
    
    def create_task_from_factory_id(self, guess, factory_id):
        window = self.create_window()
        log.debug("  window=%s" % str(window))
        task = self.create_task(factory_id)
        self.add_task_to_window(window, task)
        window.open()
        task.new(guess)
        return task
    
    def create_task_in_window(self, task_id, window):
        log.debug("creating %s task" % task_id)
        task = self.create_task(task_id)
        self.add_task_to_window(window, task)
        return task
    
    def add_task_to_window(self, window, task):
        window.add_task(task)
        window.activate_task(task)
        self.restore_perspective(window, task)
    
    def find_active_task_of_type(self, task_id):
        # Until remove_task bug is fixed, don't create any new windows, just
        # add a new task to the current window unless the task already exists
        w = list(self.windows)
        if not w:
            # OS X might not have any windows open; a menubar is allowed to
            # exist without windows.
            return None
        try:
            i = w.index(self.active_window)
            w[0:0] = [self.active_window]
            w.pop(i)
        except ValueError:
            pass
        
        for window in w:
            for t in window.tasks:
                if t.id == task_id:
                    log.debug("found non-active task in current window; activating!")
                    window.activate_task(t)
                    return t
        if window:
            task = self.create_task_in_window(task_id, window)
            return task
#        # Check active window first, then other windows
#        w = list(self.windows)
#        try:
#            i = w.index(self.active_window)
#            w[0:0] = [self.active_window]
#            w.pop(i)
#        except ValueError:
#            pass
#        
#        for window in w:
#            log.debug("window: %s" % window)
#            log.debug("  active task: %s" % window.active_task)
#            if window.active_task.id == task_id:
#                log.debug("  found active task")
#                return window.active_task
#        log.debug("  no active task matches %s" % task_id)
#        for window in w:
#            task = window.active_task
#            if task is None:
#                continue
#            # if no editors in the task, replace the task with the new task
#            log.debug("  window %s: %d" % (window, len(task.editor_area.editors)))
#            if len(task.editor_area.editors) == 0:
#                log.debug("  replacing unused task!")
#                # The bugs in remove_task seem to have been fixed so that the
#                # subsequent adding of a new task does seem to work now.  But
#                # I'm leaving in the workaround for now of simply closing the
#                # active window, forcing the new task to open in a new window.
#                if True:
#                    log.debug("removing task %s" % task)
#                    print window
#                    #window.remove_task(task)
#                    task = self.create_task_in_window(task_id, window)
#                    return task
#                else:
#                    window.close()
#                    return None

    def find_or_create_task_of_type(self, task_id):
        task = self.find_active_task_of_type(task_id)
        if not task:
            log.debug("task %s not found in active windows; creating new window" % task_id)
            window = self.create_window()
            task = self.create_task_in_window(task_id, window)
            window.open()
        return task

    # Override the default window closing event handlers only on Mac because
    # Mac allows the application to remain open while no windows are open
    if sys.platform == "darwin":
        def _on_window_closing(self, window, trait_name, event):
            # Event notification.
            self.window_closing = window_event = VetoableTaskWindowEvent(
                window=window)

            if window_event.veto:
                event.veto = True
            else:
                # Store the layout of the window.
                window_layout = window.get_window_layout()
                self._state.push_window_layout(window_layout)

        def _on_window_closed(self, window, trait_name, event):
            self.windows.remove(window)

            # Event notification.
            self.window_closed = TaskWindowEvent(window=window)

            # Was this the last window?
            if len(self.windows) == 0 and self._explicit_exit:
                self.stop()

    def _initialize_application_home(self):
        """Override the envisage.application method to force the use of standard
        config directory location instead of ~/.enthought 
        """

        from omnivore.third_party.appdirs import user_config_dir, user_log_dir, user_cache_dir
        dirname = user_config_dir(self.name)
        ETSConfig.application_home = dirname

        # Make sure it exists!
        if not os.path.exists(ETSConfig.application_home):
            os.makedirs(ETSConfig.application_home)

        dirname = user_log_dir(self.name)
        if not os.path.exists(dirname):
            os.makedirs(dirname)
        self.log_dir = dirname
        
        self.log_file_ext = "-%s" % datetime.now().strftime("%Y%m%d-%H%M%S")

        dirname = user_cache_dir(self.name)
        if not os.path.exists(dirname):
            os.makedirs(dirname)
        self.cache_dir = dirname
        
        # Prevent py2exe from creating a dialog box on exit saying that there
        # are error messages.  It thinks that anything written to stderr is an
        # error, and the python logging module redirects everything to stderr.
        # Instead, redirect stderr to a log file in the user log directory
        frozen = getattr(sys, 'frozen', False)
        if frozen in ('dll', 'windows_exe', 'console_exe'):
            # redirect py2exe stderr/stdout to log file
            log = self.get_log_file_name("py2exe")
            oldlog = sys.stdout
            sys.stdout = open(log, 'w')
            if hasattr(oldlog, "saved_text"):
                sys.stdout.write("".join(oldlog.saved_text))
            sys.stderr = sys.stdout
            
            # The logging module won't redirect to the new stderr without help
            handler = logging.StreamHandler(sys.stderr)
            logger = logging.getLogger('')
            logger.addHandler(handler)
        else:
            log = self.get_log_file_name("log")
            handler = logging.FileHandler(log)
            formatter = logging.Formatter("%(levelname)s:%(name)s:%(msg)s")
            handler.setFormatter(formatter)
            logger = logging.getLogger('')
            logger.addHandler(handler)
    
    #### Convenience methods
    
    def add_document(self, document):
        """Add document to the application list of open documents
        
        FIXME: check for duplicates?
        """
        existing = self.get_document(document.document_id)
        if existing:
            return existing
        
        document.document_id = self.next_document_id
        self.next_document_id += 1
        self.documents.append(document)
        return document
    
    def get_document(self, document_id):
        """Find an existing document given the document_id
        """
        for doc in self.documents:
            if doc.document_id == document_id:
                return doc
        return None
    
    def get_plugin_data(self, plugin_id):
        return self.plugin_data[plugin_id]
    
    def get_preferences(self, helper_object, debug=True):
        """Get preferences for a particular PreferenceHelper object.
        
        Handle mistakes in preference files by using the default value for any
        bad preference values.
        """
        try:
            helper = helper_object(preferences=self.preferences)
        except TraitError:
            # Create an empty preference object and helper so we can try
            # preferences one-by-one to see which are bad
            empty = Preferences()
            helper = helper_object(preferences=empty)
            if debug:
                log.debug("Application preferences before determining error:")
                self.preferences.dump()
            for t in helper.trait_names():
                if helper._is_preference_trait(t):
                    pref_name = "%s.%s" % (helper.preferences_path, t)
                    text_value = self.preferences.get(pref_name)
                    if text_value is None:
                        # None means the preference isn't specified, which
                        # isn't an error.
                        continue
                    try:
                        empty.set(pref_name, self.preferences.get(pref_name))
                    except:
                        log.error("Invalid preference for %s: %s. Using default value %s" % (pref_name, self.preferences.get(pref_name), getattr(helper, t)))
                        self.preferences.remove(pref_name)
                        # Also remove from default scope
                        self.preferences.remove("default/%s" % pref_name)
            if debug:
                log.debug("Application preferences after removing bad preferences:")
                self.preferences.dump()
            helper = helper_object(preferences=self.preferences)
        return helper
    
    def get_log_file_name(self, log_file_name_base, ext=""):
        filename = log_file_name_base + self.log_file_ext
        if ext:
            if not ext.startswith("."):
                filename += "."
            filename += ext
        else:
            filename += ".log"
        filename = os.path.join(self.log_dir, filename)
        return filename
    
    def save_log(self, text, log_file_name_base, ext=""):
        filename = self.get_log_file_name(log_file_name_base, ext)
        
        try:
            with open(filename, "wb") as fh:
                fh.write(text)
        except IOError:
            log.error("Failed writing %s to %s" % (log_file_name_base, filename))
    
    def get_config_dir_filename(self, subdir, json_name):
        config_dir = ETSConfig.application_home
        dirname = os.path.join(config_dir, subdir)
        if not os.path.exists(dirname):
            os.makedirs(dirname)
        return os.path.join(dirname, json_name)
    
    def get_json_data(self, json_name, default_on_error=None):
        try:
            file_path = self.get_config_dir_filename("json", json_name)
            with open(file_path, "r") as fh:
                raw = fh.read()
            json_data = jsonpickle.decode(raw)
            try:
                # new format is a list with a format identifier as the first entry
                if json_data[0] == "format=v2":
                    decoded = json_data[1]
            except KeyError:
                # deprecated format was a dictionary, and was creating an extra
                # layer of indirection by encoding the jsonpickle as another string
                json_data = json.loads(raw)
                encoded = json_data[json_name]
                decoded = jsonpickle.decode(encoded)
            return decoded
        except IOError:
            # file not found
            return default_on_error
        except ValueError:
            # bad JSON format
            log.error("Bad JSON format in preferences file: %s" % json_name)
            return default_on_error
    
    def save_json_data(self, json_name, data):
        file_path = self.get_config_dir_filename("json", json_name)
        json_data = ["format=v2", data]
        encoded = jsonpickle.encode(json_data)
        with open(file_path, "w") as fh:
            fh.write(encoded)
    
    def get_bson_data(self, bson_name):
        import bson
        
        file_path = self.get_config_dir_filename("bson", bson_name)
        with open(file_path, "r") as fh:
            raw = fh.read()
        if len(raw) > 0:
            bson_data = bson.loads(raw)
            data = bson_data[bson_name]
        else:
            raise IOError("Blank BSON data")
        return data
    
    def save_bson_data(self, bson_name, data):
        import bson
        
        file_path = self.get_config_dir_filename("bson", bson_name)
        bson_data = {bson_name: data}
        raw = bson.dumps(bson_data)
        with open(file_path, "w") as fh:
            fh.write(raw)
    
    # class attributes
    
    perspectives_loaded = False
    
    def init_perspectives(self):
        if self.perspectives_loaded:
            return
        try:
            data = self.get_json_data("perspectives")
        except IOError:
            # file not found
            data = {}
        self.perspectives = data.get("perspectives", {})
        self.last_window_size = tuple(data.get("window_size", self.default_window_size))
        log.debug("init_perspectives: size=%s" % str(self.last_window_size))
    
    def remember_perspectives(self, window):
        layout = window.get_layout()
        if layout is not None:
            p = layout.perspective
            # make sure central pane exists, otherwise will restore without
            # central pane visible
            if "name=Central;" in p:
                log.debug("remember_perspective: %s, size=%s" % (layout.id, str(window.size)))
                self.perspectives[layout.id] = p
                self.last_window_size = window.size
                data = {"perspectives": dict(self.perspectives),
                        "window_size": list(self.last_window_size),
                        }
                self.save_json_data("perspectives", data)

    def restore_perspective(self, window, task):
        # get layout object that can be used to restore perspective since we
        # only serialize the perspective string and not the whole layout object
        layout = window.get_layout()
        if task.id in self.perspectives:
            layout.perspective = self.perspectives[task.id]
            window.set_layout(layout)
        task.restore_toolbars(window)

def setup_frozen_logging():
    # set up early py2exe logging redirection, saving any messages until the log
    # file directory can be determined after the application is initialized.
    frozen = getattr(sys, 'frozen', False)
    if frozen in ('dll', 'windows_exe', 'console_exe'):
        class Blackhole(object):
            softspace = 0
            saved_text = []
            def write(self, text):
                self.saved_text.append(text)
            def flush(self):
                pass
        sys.stdout = Blackhole()
        sys.stderr = sys.stdout

def run(plugins=[], use_eggs=True, egg_path=[], image_path=[], startup_task="", application_name="", debug_log=False):
    """Start the application
    
    :param plugins: list of user plugins
    :param use_eggs Boolean: search for setuptools plugins and plugins in local eggs?
    :param egg_path: list of user-specified paths to search for more plugins
    :param startup_task string: task factory identifier for task shown in initial window
    :param application_name string: change application name instead of default Omnivore
    """
    # Enthought library imports.
    from envisage.api import PluginManager
    from envisage.core_plugin import CorePlugin
    
    # Local imports.
    from omnivore.framework.plugin import OmnivoreTasksPlugin, OmnivoreMainPlugin
    from omnivore.file_type.plugin import FileTypePlugin
    from omnivore import get_image_path
    from omnivore.utils.jobs import get_global_job_manager
    
    # Include standard plugins
    core_plugins = [ CorePlugin(), OmnivoreTasksPlugin(), OmnivoreMainPlugin(), FileTypePlugin() ]
    if sys.platform == "darwin":
        from omnivore.framework.osx_plugin import OSXMenuBarPlugin
        core_plugins.append(OSXMenuBarPlugin())
    
    import omnivore.file_type.recognizers
    core_plugins.extend(omnivore.file_type.recognizers.plugins)
    
    import omnivore.plugins
    core_plugins.extend(omnivore.plugins.plugins)
    
    # Add the user's plugins
    core_plugins.extend(plugins)
    
    # Check basic command line args
    default_parser = argparse.ArgumentParser(description="Default Parser")
    default_parser.add_argument("--no-eggs", dest="use_eggs", action="store_false", default=True, help="Do not load plugins from python eggs")
    options, extra_args = default_parser.parse_known_args()

    # The default is to use the specified plugins as well as any found
    # through setuptools and any local eggs (if an egg_path is specified).
    # Egg/setuptool plugin searching is turned off by the use_eggs parameter.
    default = PluginManager(
        plugins = core_plugins,
    )
    if use_eggs and options.use_eggs:
        from pkg_resources import Environment, working_set
        from envisage.api import EggPluginManager
        from envisage.composite_plugin_manager import CompositePluginManager
        
        # Find all additional eggs and add them to the working set
        environment = Environment(egg_path)
        distributions, errors = working_set.find_plugins(environment)
        if len(errors) > 0:
            raise SystemError('cannot add eggs %s' % errors)
        logger = logging.getLogger()
        logger.debug('added eggs %s' % distributions)
        map(working_set.add, distributions)

        # The plugin manager specifies which eggs to include and ignores all others
        egg = EggPluginManager(
            include = [
                'omnivore.tasks',
            ]
        )
        
        plugin_manager = CompositePluginManager(
            plugin_managers=[default, egg]
        )
    else:
        plugin_manager = default

    # Add omnivore icons after all image paths to allow user icon themes to take
    # precidence
    from pyface.resource_manager import resource_manager
    import os
    image_paths = image_path[:]
    image_paths.append(get_image_path("icons"))
    resource_manager.extra_paths.extend(image_paths)

    kwargs = {}
    if startup_task:
        kwargs['startup_task'] = startup_task
    if application_name:
        kwargs['name'] = application_name
    app = FrameworkApplication(plugin_manager=plugin_manager, command_line_args=extra_args, **kwargs)
    
    # Create a debugging log
    if debug_log:
        filename = app.get_log_file_name("debug")
        handler = logging.FileHandler(filename)
        logger = logging.getLogger('')
        logger.addHandler(handler)
        logger.setLevel(logging.DEBUG)
    
    # Turn off omnivore log debug messages by default
    log = logging.getLogger("omnivore")
    log.setLevel(logging.INFO)

    app.run()
    
    job_manager = get_global_job_manager()
    if job_manager is not None:
        job_manager.shutdown()



#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought naming package component>
#------------------------------------------------------------------------------
""" Manages naming contexts. Supports non-string data types and scoped
    preferences. Part of the AppTools project of the Enthought Tool Suite.

"""
from api import *



import os
import sys
import pickle

import numpy as np
import wx
import wx.lib.filebrowsebutton as filebrowse
from wx.lib.expando import ExpandoTextCtrl, EVT_ETC_LAYOUT_NEEDED

from omnivore.utils.processutil import which
from omnivore.utils.textutil import text_to_int
from omnivore.utils.wx.dropscroller import ReorderableList, PickledDropTarget, PickledDataObject


class SimplePromptDialog(wx.TextEntryDialog):
    """Simple subclass of wx.TextEntryDialog to convert text result to a
    specific format
    """
    def convert_text(self, text, return_error=False, **kwargs):
        if return_error:
            return text, ""
        else:
            return text

    def show_and_get_value(self, return_error=False, **kwargs):
        result = self.ShowModal()
        if result == wx.ID_OK:
            text = self.GetValue()
            value = self.convert_text(text, return_error, **kwargs)
        else:
            if return_error:
                value = None, "Cancelled"
            else:
                value = None
        self.Destroy()
        return value

class HexEntryDialog(SimplePromptDialog):
    """Simple subclass of wx.TextEntryDialog to convert text result from
    hexidecimal if necessary.
    """
    def convert_text(self, text, return_error=False, default_base="dec", **kwargs):
        try:
            value = text_to_int(text, default_base)
            error = ""
        except (ValueError, TypeError), e:
            value = None
            error = str(e)
        if return_error:
            return value, error
        else:
            return value

def prompt_for_string(parent, message, title, default=None):
    if default is not None:
        default = str(default)
    else:
        default = ""
    d = SimplePromptDialog(parent, message, title, default)
    return d.show_and_get_value()

def prompt_for_hex(parent, message, title, default=None, return_error=False, default_base="dec"):
    if default is not None:
        if default_base == "hex":
            default = hex(int(default))[2:]
        else:
            default = str(int(default))
    else:
        default = ""
    d = HexEntryDialog(parent, message, title, default)
    if default:
        d.SetValue(default)
    return d.show_and_get_value(return_error=return_error, default_base=default_base)


class DictEditDialog(wx.Dialog):
    border = 5
    
    def __init__(self, parent, title, instructions, fields, default=None):
        wx.Dialog.__init__(self, parent, -1, title)
        sizer = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(sizer)
        
        t = wx.StaticText(self, -1, instructions)
        sizer.Add(t, 0, wx.ALL|wx.EXPAND, self.border)
        
        self.types = {}
        self.controls = {}
        self.buttons = {}
        self.add_fields(fields)
        
        btnsizer = wx.StdDialogButtonSizer()
        self.ok_btn = wx.Button(self, wx.ID_OK)
        self.ok_btn.SetDefault()
        btnsizer.AddButton(self.ok_btn)
        btn = wx.Button(self, wx.ID_CANCEL)
        btnsizer.AddButton(btn)
        btnsizer.Realize()
        sizer.Add(btnsizer, 1, wx.ALL|wx.EXPAND, self.border)
        
        self.Bind(wx.EVT_BUTTON, self.on_button)
        self.Bind(wx.EVT_TEXT, self.on_text_changed)
        self.Bind(EVT_ETC_LAYOUT_NEEDED, self.on_resize)
        
        # Don't call self.Fit() otherwise the dialog buttons are zero height
        sizer.Fit(self)
        
        self.default = default
        if default:
            self.set_initial_data(default)
        self.check_enable()
    
    def add_fields(self, fields):
        """ Calls the create_X method where X is the type of the field. Fields
        are a list of tuples: (type, key, label) or (type, key, label, choices)
        """
        for field in fields:
            try:
                type, key, label = field
                choices = None
            except ValueError:
                type, key, label, choices = field
            try:
                func = getattr(self, "create_%s" % type.replace(" ", "_"))
            except AttributeError:
                raise NotImplementedError("Unknown field type %s for %s" % (type, key))
            control = func(key, label)
            if key is not None:
                self.types[key] = type
                self.controls[key] = control

    def set_initial_data(self, d):
        for key, control in self.controls.iteritems():
            type = self.types[key]
            if type == 'text' or type == 'verify':
                value = self.get_default_value(d, key)
                control.ChangeValue(value)
            elif type == 'verify list':
                value = "\n".join(self.get_default_value(d, key))
                control.ChangeValue(value)
            elif type == 'file' or type == 'boolean':
                value = self.get_default_value(d, key)
                control.SetValue(value)
    
    def get_default_value(self, d, key):
        return d[key]
    
    def get_edited_values(self, d):
        for key, control in self.controls.iteritems():
            type = self.types[key]
            if type == 'text' or type == 'file' or type == 'boolean':
                self.set_output_value(d, key, control.GetValue())
    
    def set_output_value(self, d, key, value):
        d[key] = value
    
    def create_text(self, key, label):
        sizer = self.GetSizer()
        hbox = wx.BoxSizer(wx.HORIZONTAL)
        t = wx.StaticText(self, -1, label)
        hbox.Add(t, 0, wx.LEFT|wx.ALIGN_CENTER_VERTICAL, self.border)
        entry = wx.TextCtrl(self, -1, size=(-1, -1))
        hbox.Add(entry, 1, wx.ALL|wx.EXPAND, self.border)
        sizer.Add(hbox, 0, wx.LEFT|wx.RIGHT|wx.TOP|wx.EXPAND, self.border)
        return entry

    def create_boolean(self, key, label):
        sizer = self.GetSizer()
        hbox = wx.BoxSizer(wx.HORIZONTAL)
        t = wx.StaticText(self, -1, label)
        hbox.Add(t, 0, wx.LEFT|wx.ALIGN_CENTER_VERTICAL, self.border)
        entry = wx.CheckBox(self, -1, size=(-1, -1))
        hbox.Add(entry, 1, wx.ALL|wx.EXPAND, self.border)
        sizer.Add(hbox, 0, wx.LEFT|wx.RIGHT|wx.TOP|wx.EXPAND, self.border)
        return entry

    def create_expando(self, key, label):
        sizer = self.GetSizer()
        status = ExpandoTextCtrl(self, style=wx.ALIGN_LEFT|wx.TE_READONLY|wx.NO_BORDER)
        attr = self.GetDefaultAttributes()
        status.SetBackgroundColour(attr.colBg)
        sizer.Add(status, 1, wx.ALL|wx.EXPAND, self.border)
        return status

    def create_file(self, key, label):
        sizer = self.GetSizer()
        hbox = wx.BoxSizer(wx.HORIZONTAL)
        entry = filebrowse.FileBrowseButton(self, -1, size=(450, -1), labelText=label, changeCallback=self.on_path_changed)
        hbox.Add(entry, 0, wx.LEFT, self.border)
        sizer.Add(hbox, 0, wx.ALL|wx.EXPAND, 0)
        return entry
    
    def create_verify(self, key, label):
        sizer = self.GetSizer()
        hbox = wx.BoxSizer(wx.HORIZONTAL)
        t = wx.StaticText(self, -1, label)
        hbox.Add(t, 0, wx.LEFT|wx.ALIGN_CENTER_VERTICAL, self.border)
        entry = wx.TextCtrl(self, -1, size=(-1, -1))
        hbox.Add(entry, 1, wx.ALL|wx.EXPAND, self.border)
        verify = wx.Button(self, -1, "Verify")
        hbox.Add(verify, 0, wx.LEFT|wx.RIGHT|wx.TOP|wx.EXPAND|wx.ALIGN_CENTER, self.border)
        sizer.Add(hbox, 0, wx.LEFT|wx.RIGHT|wx.TOP|wx.EXPAND, self.border)
        verify.Bind(wx.EVT_BUTTON, self.on_verify)
        self.buttons[key] = verify
        return entry
    
    def create_verify_list(self, key, label):
        sizer = self.GetSizer()
        hbox = wx.BoxSizer(wx.HORIZONTAL)
        t = wx.StaticText(self, -1, label)
        hbox.Add(t, 0, wx.LEFT|wx.ALIGN_CENTER_VERTICAL, self.border)
        entry = wx.TextCtrl(self, -1, size=(-1, 100), style=wx.TE_MULTILINE|wx.HSCROLL)
        hbox.Add(entry, 1, wx.ALL|wx.EXPAND, self.border)
        verify = wx.Button(self, -1, "Verify")
        hbox.Add(verify, 0, wx.LEFT|wx.RIGHT|wx.ALIGN_CENTER, self.border)
        sizer.Add(hbox, 0, wx.LEFT|wx.RIGHT|wx.TOP|wx.EXPAND, self.border)
        verify.Bind(wx.EVT_BUTTON, self.on_verify)
        self.buttons[key] = verify
        return entry

    def create_gauge(self, key, label):
        sizer = self.GetSizer()
        gauge = wx.Gauge(self, -1, 50, size=(500, 5))
        sizer.Add(gauge, 0, wx.ALL|wx.EXPAND, self.border)
        return gauge

    def create_static(self, key, label):
        sizer = self.GetSizer()
        t = wx.StaticText(self, -1, label)
        sizer.Add(t, 0, wx.LEFT|wx.RIGHT|wx.BOTTOM|wx.EXPAND, self.border)

    def create_static_spacer_below(self, key, label):
        sizer = self.GetSizer()
        t = wx.StaticText(self, -1, label)
        sizer.Add(t, 0, wx.LEFT|wx.RIGHT|wx.BOTTOM|wx.EXPAND, self.border)
        t = wx.StaticText(self, -1, " ")
        sizer.Add(t, 0, wx.ALL|wx.EXPAND, self.border)
        
    def on_button(self, evt):
        if evt.GetId() == wx.ID_OK:
            self.EndModal(wx.ID_OK)
        else:
            self.EndModal(wx.ID_CANCEL)
        evt.Skip()

    def on_text_changed(self, evt):
        self.ok_btn.Enable(self.can_submit())

    def on_path_changed(self, evt):
        pass

    def on_verify(self, evt):
        print "Verify!"

    def on_resize(self, event):
        print "resized"
        self.Fit()
    
    def can_submit(self):
        return True
    
    def check_enable(self):
        self.ok_btn.Enable(self.can_submit())

    def show_and_get_value(self):
        result = self.ShowModal()
        if result == wx.ID_OK:
            d = self.get_result_object()
            self.get_edited_values(d)
        else:
            d = None
        self.Destroy()
        return d
    
    def get_result_object(self):
        # Edit the object in place by reusing the same dictionary
        if self.default:
            d = self.default
        else:
            d = dict()
        return d

class ObjectEditDialog(DictEditDialog):
    def __init__(self, parent, title, instructions, fields, object_class, default=None):
        DictEditDialog.__init__(self, parent, title, instructions, fields, default)
        self.new_object_class = object_class
        
    def get_default_value(self, d, key):
        return getattr(d, key)
    
    def get_result_object(self):
        # Edit the object in place by reusing the same dictionary
        if self.default:
            d = self.default
        else:
            d = self.new_object_class()
        return d
    
    def set_output_value(self, d, key, value):
        setattr(d, key, value)


class EmulatorDialog(DictEditDialog):
    border = 5
    
    def __init__(self, parent, title, default=None):
        fields = [
            ('file', 'exe', 'Executable: '),
            ('text', 'args', 'Args: '),
            ('static spacer below', None, "(use %s as placeholder for the data file or it will be added at the end)"),
            ('text', 'name', 'Display Name: '),
            ]
        self.user_changed = False
        DictEditDialog.__init__(self, parent, title, "Enter emulator information:", fields, default)

    def on_text_changed(self, evt):
        if evt.GetEventObject() == self.controls['name']:
            self.user_changed = True
        elif not self.user_changed:
            self.set_automatic_name()
    
    def set_automatic_name(self):
        name = os.path.basename(self.controls['exe'].GetValue())
        args = self.controls['args'].GetValue()
        if args:
            name += " " + args
        self.controls['name'].ChangeValue(name)

    def on_path_changed(self, evt):
        if not self.user_changed:
            self.set_automatic_name()
        self.ok_btn.Enable(self.can_submit())
        
    def can_submit(self):
        path = self.controls['exe'].GetValue()
        return bool(which(path))

def prompt_for_emulator(parent, title, default_emu=None):
    d = EmulatorDialog(parent, title, default_emu)
    return d.show_and_get_value()


class AssemblerDialog(DictEditDialog):
    border = 5
    
    def __init__(self, parent, title, default=None):
        fields = [
            ('text', 'name', 'Name: '),
            ('text', 'origin', 'Origin Directive: '),
            ('text', 'data byte', 'Data Byte Directive: '),
            ('text', 'comment char', 'Comment Char: '),
            ]
        DictEditDialog.__init__(self, parent, title, "Enter assembler information:", fields, default)
    
    def can_submit(self):
        control = self.controls['name']
        return len(control.GetValue()) > 0

def prompt_for_assembler(parent, title, default=None):
    d = AssemblerDialog(parent, title, default)
    return d.show_and_get_value()


def get_file_dialog_wildcard(name, extension_list):
    # Using only the first extension
    wildcards = []
    if extension_list:
        ext = extension_list[0]
        wildcards.append("%s (*%s)|*%s" % (name, ext, ext))
    return "|".join(wildcards)


class SegmentOrderDialog(wx.Dialog):
    border = 5
    instructions = "Drag segments to the right-hand list to create the output segment order"
    dest_list_title = "Segments in Executable"
    
    def __init__(self, parent, title, segments, list_title=None, credits=False):
        wx.Dialog.__init__(self, parent, -1, title, style=wx.DEFAULT_DIALOG_STYLE|wx.RESIZE_BORDER)
        
        self.segment_map = {k:v for k,v in enumerate(segments)}
        
        sizer = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(sizer)
        
        t = wx.StaticText(self, -1, self.instructions)
        sizer.Add(t, 0, wx.ALL|wx.EXPAND, self.border)
        
        hbox = wx.BoxSizer(wx.HORIZONTAL)
        
        vbox1 = wx.BoxSizer(wx.VERTICAL)
        t = wx.StaticText(self, -1, "All Segments")
        vbox1.Add(t, 0, wx.ALL|wx.EXPAND, self.border)
        self.source = ReorderableList(self, self.segment_map.keys(), self.get_item_text, columns=["Origin", "Size", "Name"], resize_column=3, allow_drop=False, size=(400,300))
        vbox1.Add(self.source, 1, wx.ALL|wx.EXPAND, self.border)
        hbox.Add(vbox1, 1, wx.ALL|wx.EXPAND, 0)

        vbox2 = wx.BoxSizer(wx.VERTICAL)
        if list_title is None:
            list_title = self.dest_list_title
        t = wx.StaticText(self, -1, list_title)
        vbox2.Add(t, 0, wx.ALL|wx.EXPAND, self.border)
        self.dest = ReorderableList(self, [], self.get_item_text, columns=["Origin", "Size", "Name"], resize_column=3, size=(400,300))
        vbox2.Add(self.dest, 1, wx.ALL|wx.EXPAND, self.border)
        hbox.Add(vbox2, 1, wx.ALL|wx.EXPAND, 0)
        
        vbox = wx.BoxSizer(wx.VERTICAL)
        
        self.add_command_area(vbox, credits)
        
        vbox.AddStretchSpacer()
        
        btnsizer = wx.StdDialogButtonSizer()
        self.ok_btn = wx.Button(self, wx.ID_OK)
        self.ok_btn.SetDefault()
        btnsizer.AddButton(self.ok_btn)
        btn = wx.Button(self, wx.ID_CANCEL)
        btnsizer.AddButton(btn)
        btnsizer.Realize()
        vbox.Add(btnsizer, 0, wx.ALL|wx.EXPAND, self.border)
        hbox.Add(vbox, 0, wx.EXPAND, 0)
        sizer.Add(hbox, 1, wx.EXPAND, 0)
        
        self.Bind(wx.EVT_BUTTON, self.on_button)
        self.Bind(wx.EVT_TEXT, self.on_text_changed)
        
        # Don't call self.Fit() otherwise the dialog buttons are zero height
        sizer.Fit(self)
        self.check_enable()
    
    def add_command_area(self, vbox, credits):
        t = wx.StaticText(self, -1, "Run Address")
        vbox.Add(t, 0, wx.LEFT|wx.RIGHT|wx.TOP|wx.EXPAND, self.border)
        self.run_addr = wx.TextCtrl(self, -1, size=(-1, -1))
        vbox.Add(self.run_addr, 0, wx.LEFT|wx.RIGHT|wx.BOTTOM|wx.EXPAND, self.border)
        
        if credits:
            t = wx.StaticText(self, -1, "Title (20 chars)")
            vbox.Add(t, 0, wx.LEFT|wx.RIGHT|wx.TOP|wx.EXPAND, self.border)
            self.title_20 = wx.TextCtrl(self, -1, size=(-1, -1))
            self.title_20.SetMaxLength(20)
            vbox.Add(self.title_20, 0, wx.LEFT|wx.RIGHT|wx.BOTTOM|wx.EXPAND, self.border)

            t = wx.StaticText(self, -1, "Author (20 chars)")
            vbox.Add(t, 0, wx.LEFT|wx.RIGHT|wx.TOP|wx.EXPAND, self.border)
            self.author_20 = wx.TextCtrl(self, -1, size=(-1, -1))
            self.author_20.SetMaxLength(20)
            vbox.Add(self.author_20, 0, wx.LEFT|wx.RIGHT|wx.BOTTOM|wx.EXPAND, self.border)
        else:
            self.title_20 = self.author_20 = None

        vbox.AddSpacer(50)
        
        self.clear = b = wx.Button(self, wx.ID_DOWN, 'Clear List', size=(90, -1))
        b.Bind(wx.EVT_BUTTON, self.on_clear)
        vbox.Add(b, 0, wx.ALL|wx.EXPAND, self.border)
    
    def get_item_text(self, sid):
        s = self.segment_map[sid]
        return "%x" % s.start_addr, "%x" % len(s), s.name
        
    def on_button(self, evt):
        if evt.GetId() == wx.ID_OK:
            self.EndModal(wx.ID_OK)
        else:
            self.EndModal(wx.ID_CANCEL)
        evt.Skip()

    def on_text_changed(self, evt):
        self.ok_btn.Enable(self.can_submit())

    def on_clear(self, evt):
        self.dest.clear()

    def on_resize(self, event):
        print "resized"
        self.Fit()
    
    def get_run_addr(self):
        text = self.run_addr.GetValue()
        try:
            addr = text_to_int(text, "hex")
            if addr < 0 or addr > 0xffff:
                addr = None
        except (ValueError, TypeError), e:
            addr = None
        return addr
    
    def can_submit(self):
        return self.get_run_addr() is not None and self.dest.GetItemCount() > 0
    
    def check_enable(self):
        self.ok_btn.Enable(self.can_submit())
    
    def get_segments(self):
        s = []
        for sid in self.dest.items:
            s.append(self.segment_map[sid])
        return s

    def get_bytes(self):
        segments = self.get_segments()
        total = 2
        for s in segments:
            total += 4 + len(s)
        total += 6
        bytes = np.zeros([total], dtype=np.uint8)
        bytes[0:2] = 0xff # FFFF header
        i = 2
        for s in segments:
            words = bytes[i:i+4].view(dtype='<u2')
            words[0] = s.start_addr
            words[1] = s.start_addr + len(s) - 1
            i += 4
            bytes[i:i + len(s)] = s[:]
            i += len(s)
        words = bytes[i:i+6].view(dtype='<u2')
        words[0] = 0x2e0
        words[1] = 0x2e1
        words[2] = self.get_run_addr()
        return bytes

    def get_extra_text(self):
        lines = []
        if self.title_20 is not None:
            lines.append(self.title_20.GetValue())
        if self.author_20 is not None:
            lines.append(self.author_20.GetValue())
        return lines


class SegmentInterleaveDialog(SegmentOrderDialog):
    instructions = "Drag segments to the right-hand list to determine interleave order. Then, choose interleave factor in bytes."
    dest_list_title = "Interleave Order"

    def add_command_area(self, vbox):
        t = wx.StaticText(self, -1, "Interleave Bytes")
        vbox.Add(t, 0, wx.ALL|wx.EXPAND, self.border)
        self.intervleave = wx.TextCtrl(self, -1, size=(-1, -1))
        vbox.Add(self.intervleave, 0, wx.ALL|wx.EXPAND, self.border)
        
        vbox.AddSpacer(50)
        
        self.clear = b = wx.Button(self, wx.ID_DOWN, 'Clear List', size=(90, -1))
        b.Bind(wx.EVT_BUTTON, self.on_clear)
        vbox.Add(b, 0, wx.ALL|wx.EXPAND, self.border)
    
    def get_length(self):
        segments = self.get_segments()
        if len(segments) == 0:
            return -1
        common = len(segments[0])
        for s in segments[1:]:
            if len(s) != common:
                return -1
        return common
    
    def get_interleave(self):
        text = self.intervleave.GetValue()
        length = self.get_length()
        try:
            num = text_to_int(text, "hex")
            if num < 0 or num > length:
                num = 0
        except (ValueError, TypeError), e:
            num = 0
        return num
    
    def can_submit(self):
        return self.get_length() > 0 and self.get_interleave() > 0


class ListReorderDialog(wx.Dialog):
    """Simple dialog to return a list of items that can be reordered by the user.
    """
    border = 5
    
    def __init__(self, parent, items, get_item_text, dialog_helper=None, title="Reorder List", copy_helper=None):
        wx.Dialog.__init__(self, parent, -1, title,
                           size=(700, 500), pos=wx.DefaultPosition, 
                           style=wx.DEFAULT_DIALOG_STYLE|wx.RESIZE_BORDER)

        sizer = wx.BoxSizer(wx.HORIZONTAL)

        self.list = ReorderableList(self, items, get_item_text, size=(-1,500))
        self.list.Bind(wx.EVT_LIST_ITEM_SELECTED, self.on_list_selection)
        self.list.Bind(wx.EVT_LIST_ITEM_DESELECTED, self.on_list_selection)
        sizer.Add(self.list, 1, wx.EXPAND)

        vbox = wx.BoxSizer(wx.VERTICAL)
        
        self.up = b = wx.Button(self, wx.ID_UP, 'Up', size=(90, -1))
        b.Bind(wx.EVT_BUTTON, self.on_up)
        vbox.Add(b, 0, wx.ALL|wx.EXPAND, self.border)
        
        self.down = b = wx.Button(self, wx.ID_DOWN, 'Down', size=(90, -1))
        b.Bind(wx.EVT_BUTTON, self.on_down)
        vbox.Add(b, 0, wx.ALL|wx.EXPAND, self.border)
        
        vbox.AddSpacer(50)
        
        b = wx.Button(self, wx.ID_NEW, 'New', size=(90, -1))
        b.Bind(wx.EVT_BUTTON, self.on_new)
        vbox.Add(b, 0, wx.ALL|wx.EXPAND, self.border)
        
        self.edit = b = wx.Button(self, wx.ID_EDIT, 'Edit', size=(90, -1))
        b.Bind(wx.EVT_BUTTON, self.on_edit)
        vbox.Add(b, 0, wx.ALL|wx.EXPAND, self.border)
        
        if copy_helper is not None:
            self.copy = b = wx.Button(self, wx.ID_COPY, 'Copy', size=(90, -1))
            b.Bind(wx.EVT_BUTTON, self.on_copy)
            vbox.Add(b, 0, wx.ALL|wx.EXPAND, self.border)
        else:
            self.copy = None
        
        vbox.AddSpacer(50)
        
        self.delete = b = wx.Button(self, wx.ID_DELETE, 'Delete', size=(90, -1))
        b.Bind(wx.EVT_BUTTON, self.on_delete)
        vbox.Add(b, 0, wx.ALL|wx.EXPAND, self.border)
        
        vbox.AddStretchSpacer()
        
        btnsizer = wx.StdDialogButtonSizer()
        btn = wx.Button(self, wx.ID_OK)
        btn.SetDefault()
        btnsizer.AddButton(btn)
        btn = wx.Button(self, wx.ID_CANCEL)
        btnsizer.AddButton(btn)
        btnsizer.Realize()
        vbox.Add(btnsizer, 0, wx.ALL|wx.EXPAND, self.border)
        sizer.Add(vbox, 0, wx.EXPAND, 0)

        self.SetSizer(sizer)
        sizer.Fit(self)

        self.Layout()
        
        self.list.Bind(wx.EVT_CONTEXT_MENU, self.on_context_menu)
        self.delete_id = wx.NewId()
        self.Bind(wx.EVT_MENU, self.on_delete, id=self.delete_id)
        
        self.get_item_text = get_item_text
        self.dialog_helper = dialog_helper
        self.copy_helper = copy_helper
        self.on_list_selection(None)
    
    def on_list_selection(self, evt):
        one_selected = self.list.GetSelectedItemCount() == 1
        any_selected = self.list.GetSelectedItemCount() > 0
        self.up.Enable(self.list.can_move_up)
        self.down.Enable(self.list.can_move_down)
        self.edit.Enable(one_selected)
        if self.copy is not None:
            self.copy.Enable(one_selected)
        self.delete.Enable(any_selected)
    
    def on_context_menu(self, evt):
        one_selected = self.list.GetSelectedItemCount() == 1
        any_selected = self.list.GetSelectedItemCount() > 0
        menu = wx.Menu()
        menu.Append(wx.ID_NEW, "New Item")
        menu.Append(wx.ID_EDIT, "Edit Item")
        menu.Enable(wx.ID_EDIT, one_selected)
        if self.copy is not None:
            menu.Append(wx.ID_COPY, "Copy Item")
            menu.Enable(wx.ID_COPY, one_selected)
        menu.AppendSeparator()
        menu.Append(wx.ID_SELECTALL, "Select All")
        menu.Append(wx.ID_CLEAR, "Deselect All")
        menu.Append(wx.ID_DELETE, "Delete Selected Items")
        menu.Enable(wx.ID_DELETE, any_selected)
        id = self.GetPopupMenuSelectionFromUser(menu)
        menu.Destroy()
        if id == wx.ID_NEW:
            self.on_new(evt)
        elif id == wx.ID_DELETE:
            self.on_delete(evt)
        elif id == wx.ID_EDIT:
            self.on_edit(evt)
        elif id == wx.ID_COPY:
            self.on_copy(evt)
        if id == wx.ID_SELECTALL:
            self.list.select_all()
        elif id == wx.ID_CLEAR:
            self.list.deselect_all()

    def get_items(self):
        return self.list.items
    
    def on_up(self, evt):
        if self.list.can_move_up:
            self.list.move_selected(-1)
    
    def on_down(self, evt):
        if self.list.can_move_down:
            self.list.move_selected(1)
    
    def on_new(self, evt):
        new_item = self.dialog_helper(self, "Add Item")
        if new_item is not None:
            self.insert_new_item(new_item)
    
    def insert_new_item(self, new_item):
        index = self.list.GetFirstSelected()
        if index == -1:
            index = len(self.list.items)
        else:
            index += 1
        self.list.items[index:index] = [new_item]
        self.list.refresh()
        self.list.deselect_all()
        self.list.SetItemState(index, wx.LIST_STATE_SELECTED, wx.LIST_STATE_SELECTED)
    
    def on_edit(self, evt):
        index = self.list.GetFirstSelected()
        if index >= 0:
            item = self.list.items[index]
            new_item = self.dialog_helper(self, "Edit %s" % self.get_item_text(item), item)
            if new_item is not None:
                self.list.items[index] = new_item
                self.list.refresh()
    
    def on_copy(self, evt):
        index = self.list.GetFirstSelected()
        if index >= 0:
            item = self.list.items[index]
            new_item = self.copy_helper(item)
            if new_item is not None:
                self.insert_new_item(new_item)
    
    def on_delete(self, evt):
        self.list.delete_selected()


class CheckItemDialog(wx.Dialog):
    """Simple dialog to return a list of items that can be reordered by the user.
    """
    border = 5
    
    def __init__(self, parent, items, get_item_text, dialog_helper=None, title="Select Items", instructions=""):
        wx.Dialog.__init__(self, parent, -1, title,
                           size=(700, 500), pos=wx.DefaultPosition, 
                           style=wx.DEFAULT_DIALOG_STYLE|wx.RESIZE_BORDER)

        sizer = wx.BoxSizer(wx.VERTICAL)

        self.get_item_text = get_item_text
        self.dialog_helper = dialog_helper

        if instructions:
            t = wx.StaticText(self, -1, instructions)
            sizer.Add(t, 0, wx.ALL|wx.ALIGN_CENTER_VERTICAL, self.border)

        # Use multi-select to make sure no initial item is selecte. Single
        # select forces the first item to be selected.
        self.list = wx.CheckListBox(self, size=(-1,500), style=wx.LB_MULTIPLE)
        
        # Both EVT_LISTBOX and EVT_CHECKLISTBOX cancel each other out when
        # clicking on the check box itself. Clicking only on item text causes
        # EVT_LISTBOX event only.
#        self.list.Bind(wx.EVT_LISTBOX, self.on_list_selection)
        self.list.Bind(wx.EVT_CHECKLISTBOX, self.on_list_check)
        self.list.Bind(wx.EVT_CONTEXT_MENU, self.on_context_menu)
        sizer.Add(self.list, 1, wx.EXPAND)
        self.set_items(items)
        
        btnsizer = wx.StdDialogButtonSizer()
        btn = wx.Button(self, wx.ID_OK)
        btn.SetDefault()
        btnsizer.AddButton(btn)
        btn = wx.Button(self, wx.ID_CANCEL)
        btnsizer.AddButton(btn)
        btnsizer.Realize()
        sizer.Add(btnsizer, 0, wx.EXPAND, 0)

        self.SetSizer(sizer)
        sizer.Fit(self)

        self.Layout()
        
        self.Bind(wx.EVT_CONTEXT_MENU, self.on_context_menu)
        self.delete_id = wx.NewId()
    
    def set_items(self, items):
        self.clear()
        for item in items:
            self.insert_item(self.list.GetCount(), item)
    
    def insert_item(self, index, item):
        self.list.Insert("placeholder", index)
        self.set_item_text(index, item)
        self.items[index:index] = [item]
    
    def set_item_text(self, index, item):
        text = self.get_item_text(item)
        self.list.SetString(index, text)
    
    def on_list_selection(self, evt):
        index = evt.GetInt()
        self.list.Check(index, not self.list.IsChecked(index))
        self.list.SetSelection(index)
    
    def on_list_check(self, evt):
        index = evt.GetInt()
    
    def on_context_menu(self, evt):
        menu = wx.Menu()
        menu.Append(wx.ID_SELECTALL, "Select All")
        menu.Append(wx.ID_CLEAR, "Deselect All")
        id = self.GetPopupMenuSelectionFromUser(menu)
        menu.Destroy()
        if id == wx.ID_SELECTALL:
            self.select_all()
        elif id == wx.ID_CLEAR:
            self.deselect_all()

    def select_all(self):
        self.list.SetChecked(range(self.list.GetCount()))
    
    def deselect_all(self):
        self.list.SetChecked([])

    def get_items(self):
        return self.items
    
    def get_checked_items(self):
        return [self.items[i] for i in self.list.GetChecked()]
    
    def clear(self, evt=None):
        self.list.Clear()
        self.items = []


if __name__ == "__main__":
    app = wx.PySimpleApp()

    frame = wx.Frame(None, -1, "Dialog test")
#    dialog = EmulatorDialog(frame, "Test")
#    dialog.ShowModal()
#    dlg = ListReorderDialog(frame, [chr(i + 65) for i in range(26)], lambda a:str(a))
    dlg = CheckItemDialog(frame, [chr(i + 65) for i in range(26)], lambda a:str(a))
    if dlg.ShowModal() == wx.ID_OK:
        print dlg.get_checked_items()
    dlg.Destroy()

    app.MainLoop()



from __future__ import absolute_import

import os
import sys

from traits.testing.unittest_tools import unittest, UnittestTools

from ..gui import GUI
from ..python_editor import PythonEditor
from ..window import Window


PYTHON_SCRIPT = os.path.join(os.path.dirname(__file__), 'python_shell_script.py')


class TestPythonEditor(unittest.TestCase, UnittestTools):

    def setUp(self):
        self.gui = GUI()
        self.window = Window()
        self.window._create()

    def tearDown(self):
        self.widget.destroy()
        self.window.destroy()

    def test_lifecycle(self):
        # test that destroy works
        self.widget = PythonEditor(self.window.control)
        self.gui.process_events()
        self.assertFalse(self.widget.dirty)
        self.widget.destroy()
        self.gui.process_events()

    def test_show_line_numbers(self):
        # test that destroy works
        self.widget = PythonEditor(self.window.control, show_line_numbers=False)
        self.gui.process_events()
        self.widget.show_line_numbers = True
        self.gui.process_events()
        self.widget.show_line_numbers = False
        self.gui.process_events()
        self.widget.destroy()
        self.gui.process_events()

    def test_load(self):
        # test that destroy works
        self.widget = PythonEditor(self.window.control)
        self.gui.process_events()
        with self.assertTraitChanges(self.widget, 'changed', count=1):
            self.widget.path = PYTHON_SCRIPT
        self.assertFalse(self.widget.dirty)
        self.widget.destroy()
        self.gui.process_events()

    def test_select_line(self):
        # test that destroy works
        self.widget = PythonEditor(self.window.control, path=PYTHON_SCRIPT)
        self.gui.process_events()
        self.widget.select_line(3)
        self.widget.destroy()
        self.gui.process_events()
